/*
	Declarations section
*/
%{
#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/atyacc.y,v 2.10 2000/04/14 01:17:54 johnh Exp $";
#endif

/*
 * $Log: atyacc.y,v $
 * Revision 2.10  2000/04/14 01:17:54  johnh
 * patch for y2k problem in at jobs
 *
 * Revision 2.9  2000/03/23 01:33:24  geoff
 * Fix the year-handling code to assume that 2-digit years are within the
 * current century.
 *
 * Revision 2.8  1999/07/27 01:22:00  johnh
 * location support for at jobs
 *
 * Revision 2.7  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.6  1995/11/27 01:59:22  geoff
 * Support relative day specifications.
 *
 * Revision 2.5  1995/05/11  05:05:01  geoff
 * Minor RCS header changes.
 *
 * Revision 2.4  88/10/31  23:53:01  geoff
 * Add code to distinguish 1- and 2-digit times from 3- and 4-digit ones,
 * so that we can be smart about recognizing things like "at 11pm".
 * 
 * Revision 2.3  86/03/25  19:55:06  geoff
 * Add zulu detection, and fix a bug in hour:minute parsing.
 * 
 * Revision 2.2  86/01/19  13:35:53  geoff
 * Add actions to set the time parsed into 'when'.
 * 
 * Revision 2.1  86/01/04  20:05:47  geoff
 * Initial revision
 * 
 */

#include <stdio.h>
#include <time.h>
#include <sys/types.h>

typedef int		yacctype;	/* Yacc stack type */
#define YYSTYPE		yacctype	/* Yacc stack type */

static int		numsize;	/* Number of digits in a number */

extern int		incrcode;	/* Code for incrementation unit */
extern int		incrval;	/* Number of units to increment by */
extern struct tm *	now;		/* What time it is now */
extern struct tm	when;		/* When to run the command */
/* #ifdef LOC_CRON */
/* #ifdef LOC_ATLOC */
char *			where;		/* where to run the command */
/* #endif \* LOC_ATLOC */
/* #endif \* LOC_CRON */
extern int		zulu;		/* NZ if Zulu (GMT) */
%}
/*
 * Keyword definitions
 */

%token K_AM_PM
%token K_BIG_NUMBER
%token K_INTERVAL
%token K_MIDNIGHT
%token K_MONTH
%token K_NEXT
%token K_NOON
%token K_NOW
%token K_NUMBER
%token K_RELDAY
%token K_SYNTAX
%token K_WEEKDAY
%token K_PLACE
%%
/*
	Rules section
*/

when	:	place time date increment place

place	:	/* empty */
	|	K_PLACE
	;

time	:	/* empty */
	|	K_NOW
		    {
		    when.tm_hour = now->tm_hour;
		    when.tm_min = now->tm_min;
		    }
	|	K_MIDNIGHT
		    {
		    when.tm_hour = when.tm_min = 0;
		    }
	|	K_NOON
		    {
		    when.tm_hour = 12;
		    when.tm_min = 0;
		    }
	|	hourmin
		    {
		    if (numsize <= 2)
			{
			when.tm_hour = $1;
			when.tm_min = 0;
			}
		    else
			{
			when.tm_hour = $1 / 100;
			when.tm_min = $1 % 100;
			}
		    }
	|	hourmin K_AM_PM
		    {
		    if (numsize <= 2)
			{
			when.tm_hour = $1;
			when.tm_min = 0;
			}
		    else
			{
			when.tm_hour = $1 / 100;
			when.tm_min = $1 % 100;
			}
		    switch ($2)
			{
			case 0:		/* AM */
			    if (when.tm_hour == 12)
				when.tm_hour = 0;
			    else if (when.tm_hour > 12)
				when.tm_hour = 99;	/* Illegal value */
			    break;
			case 1:		/* PM */
			    if (when.tm_hour <= 12)
				when.tm_hour += 12;
			    else if (when.tm_hour > 12)
				when.tm_hour = 99;	/* Illegal value */
			    break;
			case 2:		/* Zulu */
			    if (when.tm_hour == 24)
				{
				if (when.tm_min != 0)
				    when.tm_hour = 99;	/* Illegal value */
				else
				    when.tm_hour = 0;
				}
			    zulu = 1;			/* Mark Greenwich */
			    break;
			}
		    }
	;

hourmin	:	K_NUMBER
		    {
		    numsize = 2;
		    }
	|	K_BIG_NUMBER
		    {
		    numsize = 4;
		    }
	|	K_NUMBER ':' K_NUMBER
		    {
		    numsize = 4;	/* Always use hour/min interpretation */
		    if ($1 < 100  &&  $3 < 100)
			$$ = $1 * 100 + $3;
		    else
			$$ = 99;	/* Illegal value to produce error */
		    }
	;

date	:	/* Empty */
	|	K_RELDAY
		    {
		    if ($1 == 0)
			{
			/* Today */
			when.tm_mon = now->tm_mon;
			when.tm_mday = now->tm_mday;
			when.tm_year = now->tm_year;
			}
		    else
			{
			/*
			 * We handle 'tomorrow' by just setting the
			 * date to the next weekday.  The code in
			 * fixwhen will do the rest for us.
			 */
			when.tm_wday = (now->tm_wday + 1) % 7;
			}
		    }
	|	K_WEEKDAY
		    {
		    when.tm_wday = $1;
		    }
	|	K_MONTH K_NUMBER
		    {
		    when.tm_mon = $1 - 1;
		    when.tm_mday = $2;
		    }
	|	K_MONTH K_NUMBER year		/* Jan 1 1986 */
		    {
		    /* convert date to clib-style (0-based month, 1900-based year) */
		    when.tm_mon = $1 - 1;
		    when.tm_mday = $2;
		    when.tm_year = $3;
		    if (when.tm_year >= 100)
			when.tm_year -= 1900;
		    else
			{
			/*
			 * The year was given as two digits.  Assume
			 * that the user meant for the year to be
			 * within the current century.
			 */
			when.tm_year += ((int)(now->tm_year / 100)) * 100;
			}
		    }
	|	K_MONTH K_NUMBER ',' year	/* Jan 1, 1986 */
		    {
		    /* convert date to clib-style (0-based month, 1900-based year) */
		    when.tm_mon = $1 - 1;
		    when.tm_mday = $2;
		    when.tm_year = $4;
		    if (when.tm_year >= 100)
			when.tm_year -= 1900;
		    else
			{
			/*
			 * The year was given as two digits.  Assume
			 * that the user meant for the year to be
			 * within the current century.
			 */
			when.tm_year += ((int)(now->tm_year / 100)) * 100;
			}
		    }
	;

year	:	K_NUMBER
	|	K_BIG_NUMBER
	;

increment :	/* Empty */
	|	'+' K_INTERVAL
		    {
		    incrcode = $2;
		    incrval = 1;
		    }
	|	'+' K_NUMBER K_INTERVAL
		    {
		    incrcode = $3;
		    incrval = $2;
		    }
	|	'+' K_BIG_NUMBER K_INTERVAL
		    {
		    incrcode = $3;
		    incrval = $2;
		    }
	|	K_NEXT K_INTERVAL
		    {
		    incrcode = $2;
		    incrval = 1;
		    }
	;
%%
yyerror(string)
    char *		string;
    {
    fprintf(stderr, "at: %s\n", string);
    }
