#ifndef lint
static char Rcs_Id[] =
	"$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/batch/batch.c,v 1.6 1995/11/27 05:03:53 geoff Exp $";
#endif

/*
 *  batch [-v] [-Pn] queue [file ...]
 */

/*
 * $Log: batch.c,v $
 * Revision 1.6  1995/11/27 05:03:53  geoff
 * Get rid of some warning messages.
 *
 * Revision 1.5  86/11/29  19:15:50  geoff
 * Get rid of an unpleasant warning message.
 * 
 * Revision 1.4  86/11/29  15:57:45  geoff
 * Major changes to support configurability on various systems.  Much of
 * this consists of replacing BSD ifdef's with more feature-specific ones.
 * 
 * Revision 1.3  86/07/06  23:58:37  geoff
 * Fix a whole bunch of lint complaints.
 * 
 * Revision 1.2  86/06/26  23:29:32  geoff
 * Major changes to make it portable to System V.  Also add Korn shell support.
 * 
 */

#define GEN_ID

#include "config.h"
#include <sys/param.h>
#ifndef PARAMGETSTYPES
#include <sys/types.h>
#endif
#ifdef NULL
#undef NULL
#endif
#include <stdio.h>
#ifndef NULL
/* What a bunch of garbage, all to avoid warning messages... */
#define NULL 0
#endif
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>

#ifndef MAXPATHLEN
#define MAXPATHLEN	512		/* Size of a long pathname */
#endif

char grade = 'q';	/* Priority of this job */
int vflag;
extern errno;
extern void exit ();
extern char *getenv();
#ifdef HASGETWD
extern char *getwd();
#else
char *getwd();
#endif
extern char *strchr();
extern long lseek ();
extern void perror ();
extern unsigned int sleep ();
extern long time ();

enum { BOURNE, CSH, KSH } theshell;

char *deleteerr;
char *tffmt = "%s/%.14s/tf%c%ld";
char *cffmt = "%s/%.14s/cf%c%ld";
char Pid[] = PIDFILE;

main(argc, argv, arge)
char **argv;
char **arge;
{
	int i;
	reg1 int j;
	reg2 char **ep;
	reg4 char *queue = NULL;
	char tfname[256];
	char cfname[256];
	long unique();
	long spoolnumber;
	reg5 char *shell;
	reg3 FILE *tf;
	char dir[MAXPATHLEN];
	struct stat sb;
	void cleanup();

#ifndef DEBUG
	if(geteuid() != 0) {
		(void) fprintf (stderr,
		    "batch: System error-Effective uid is %d, not root\n",
		    getuid());
		cleanup ();
	}
#endif
	/*
	 *  Find and validate the grade.  a-z, only SU can set it to values
	 *  lower than 'h'.
	 */
	for(i = 1 ; i < argc ; i++) {
		if(argv[i][0] == '-')
			switch(argv[i][1]) {
			case 'P':
				grade = argv[i][2];
				break;
			case 'v':
				vflag = 1;
				break;
			default:
				Usage();
			}
	}
	if(grade < 'a'  ||  grade > 'z'
	    ||  ((grade < 'h')  &&  getuid() != 0)) {
		(void) fprintf (stderr,
		  "batch: Grade `%c' is invalid.\n", grade);
		cleanup ();
	}
	/*
	 *  First non - arg is the queue.
	 */
	for(i = 1 ; i < argc ; i++)
		if(argv[i][0] != '-') {
			queue = argv[i];
			break;
		}
	if(queue == NULL)
		Usage();
	if(signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal (SIGINT, cleanup);
	if(signal(SIGHUP, SIG_IGN) != SIG_IGN)
		(void) signal (SIGHUP, cleanup);
	if(signal(SIGQUIT, SIG_IGN) != SIG_IGN)
		(void) signal (SIGQUIT, cleanup);
	(void) sprintf (tfname, "%s/%.14s", SPOOLDIR, queue);
	if(stat(tfname, &sb) < 0  ||
	   (errno=ENOENT, (sb.st_mode & S_IFMT) != S_IFDIR)) {
		if(errno == ENOENT) {
			(void) fprintf (stderr,
			  "batch: %s: No such queue\n", queue);
			cleanup ();
		}
		else {
			perror(tfname);
			(void) fprintf (stderr,
			  "batch: %s: System error\n", tfname);
			cleanup ();
		}
	}
	/*
	 *  Spool file name starts with tf, followed by grade, followed by
	 *  a unique sequential number.  The files are sorted by this name
	 *  for priority of execution.
	 */
	spoolnumber = unique();
	(void) sprintf (tfname, tffmt, SPOOLDIR, queue, grade, spoolnumber);
	deleteerr = tfname;
	(void) umask (0077);
	if((tf = fopen(tfname, "w")) == NULL) {
		perror(tfname);
		(void) fprintf (stderr, "batch: %s: System error\n", tfname);
		cleanup ();
	}
	/*
	 *  Since it will be directly executable by the system put out a
	 *  system magic word followed by the shell.
	 *  THESHELL tells which type of environment setting to use.
	 */
	if((shell = getenv("SHELL")) == NULL)
		shell = "/bin/sh";
	if(strcmp(shell, "/bin/sh") == 0)
		theshell = BOURNE;
	else if(strncmp(shell+strlen(shell)-3, "csh", 3) == 0)
		theshell = CSH;
	else if (strncmp (shell + strlen(shell) - 3, "ksh", 3) == 0)
		theshell = KSH;
	else {
		(void) fprintf (stderr,
		    "batch: %s: Unrecognized shell in $SHELL variable\n",
		    shell);
		cleanup ();
	}
#ifdef SHARPBANG
	(void) fprintf (tf, "%s%s\n", SBSTRING, shell);
#else
	(void) fprintf (tf,
	  ": Use /bin/sh\n%s <<'IhOpEtHiSiSuNiQuE!'\n", shell);
#endif
	/*
	 *  Set everything up so that it will execute in the same env. as we
	 *  have right now.  BUG::::  Since we're root, we can get the current
	 *  directory, but since the chdir back here is done by the user prog
	 *  later that could fail if he didn't have full perms to get to the
	 *  current location.   We should at very least mention this to the user
	 */
	if(getwd(dir) == NULL) {
		(void) fprintf (stderr,
		  "batch: getwd: System error: %s\n", dir);
		cleanup ();
	}
	if( theshell == CSH )
		(void) fprintf (tf, "set histchars\n");
	(void) fprintf (tf, "cd ");
	putstring(tf, dir);
	(void) fprintf (tf, "\n");
	/*
	 *  Dump the current environment variables.  Ignore them if they
	 *  don't have = signs, throw away TERMCAP and TERM.
	 */
	for(ep = arge ; *ep ; ep++) {
		reg6 char *p;

		if((p = strchr(*ep, '=')) == NULL)
			continue;
		*p++ = 0;
		if(strcmp(*ep, "TERMCAP") == 0)
			continue;
		if(strcmp(*ep, "TERM") == 0)
			continue;
		putenv(tf, *ep, p);
	}
	/*
	 *  If using csh, set $home so the ~ character expands correctly.
	 */
	if(theshell == CSH)
		{
		(void) fprintf (tf, "set histchars='!^'\n");
		(void) fprintf (tf, "set home=$HOME\n");
		}
	/*
	 *  Dump the files specified; or stdin otherwise.
	 */
	j = 0;
	for(i++ ; i < argc ; i++) {
		FILE *input;

		if(argv[i][0] == '-')
			continue;
		if((input = fopen(argv[i], "r")) == NULL) {
			perror(argv[i]);
			cleanup();
		}
		concat(tf, input);
		++j;
	}
	if(j == 0)
		concat(tf, stdin);
#ifndef SHARPBANG
	(void) fprintf (tf, "IhOpEtHiSiSuNiQuE!\n");
#endif
	/*
	 *  Set everything up for direct execution; and then link it to
	 *  its real filename.
	 */
#ifdef HASFCHOWN
	(void) fchown (fileno (tf), getuid (), getgid ());
#else
	(void) chown (tfname, getuid (), getgid ());
#endif
#ifdef HASFCHMOD
	(void) fchmod (fileno (tf), 0700);
#else
	(void) chmod (tfname, 0700);
#endif
	(void) fclose (tf);
	(void) sprintf (cfname, cffmt, SPOOLDIR, queue, grade, spoolnumber);
#ifdef HASRENAME
	if (rename(tfname, cfname) < 0) {
		perror(tfname);
		(void) fprintf (stderr, "batch: %s: System error\n", tfname);
		cleanup ();
	}
#else
	if (link (tfname, cfname) < 0  ||  unlink (tfname) < 0) {
		perror(tfname);
		(void) fprintf (stderr, "batch: %s: System error\n", tfname);
		cleanup ();
	}
#endif
	/*
	 * Send SIGALRM to the batchd to wake it up.
	 */
	if ((tf = fopen(Pid, "r")) != NULL) {
		if (fscanf(tf, "%d", &i) == 1 && i > 0)
			(void) kill(i, SIGALRM);
		(void) fclose (tf);
	}
	if(vflag)
		(void) printf ("%s\n", cfname);
	exit(0);
}

void cleanup()
{
	if(deleteerr != NULL)
		(void) unlink (deleteerr);
	exit(1);
}

concat(to, from)
	reg1 FILE *to;
	reg2 FILE *from;
{
	reg3 int c;

	while ((c = getc(from)) != EOF)
		(void) putc (c, to);
}

/*
 *  Dump out an environment variable in the format for this shell.
 */
putenv(tf, ep, p)
FILE *tf;
char *ep, *p;
{
	switch(theshell) {
	case BOURNE:
	case KSH:
		(void) fprintf (tf, "%s=", ep);
		putstring(tf, p );
		(void) fprintf (tf, "; export %s\n", ep);
		break;
	case CSH:
		(void) fprintf (tf, "setenv %s ", ep);
		putstring(tf, p );
		(void) fprintf (tf, "\n");
		break;
	}
}

/*
 *  God help us; put out a string with whatever escapes are needed to make
 *  damn sure that the shell doesn't interpret it.  BOURNE shell doesn't
 *  need escapes on newlines; CSH does.
 */
putstring(tf, s)
reg2 FILE *tf;
reg1 char *s;
{
	(void) fputc ('\'', tf);
	while(*s) {
		if(*s == '\n' && theshell == CSH )
			(void) fputc ('\\', tf);
		if(*s == '\'')
			{
			/* 'You can'"'"'t come' */
			s++;
			(void) fprintf (tf,"'\"'\"'");
			}
		else
			(void) fputc (*s++, tf);
	}
	(void) fputc ('\'', tf);
}

Usage()
{
	(void) fprintf (stderr,
	  "batch: Usage: batch [-Pn] queue [file ...]\n");
	cleanup ();
}

/*
 * BSD has this as a system call;  USG has to do it the old way.
 */
#ifndef HASGETWD
char *getwd (dest)
reg1 char *dest;
{
    reg2 int nbytes;
    FILE *pwdfile;

    *dest = '\0';
    pwdfile = popen ("pwd", "r");
    if (pwdfile == NULL)
	return NULL;
    nbytes = fread (dest, 1, MAXPATHLEN, pwdfile);
    (void) pclose (pwdfile);
    if (nbytes <= 0)
	return NULL;
    if (nbytes > 0  &&  dest[nbytes - 1] == '\n')
	nbytes--;
    dest[nbytes] = '\0';
    return dest;
}
#endif

#ifndef HASUNIQUE
static void lockit ();
static void unlockit ();

/*
 * I guess this must be a BSD system call too, though I sure don't
 * remember it offhand.  That's ok;  we're tough.
 */
long unique ()				/* Return unique integer */
{
    char		lockpath[sizeof SPOOLDIR + 20];
    reg1 int		ufd;		/* Fd to access .uniqueval */
    char		uniquepath[sizeof SPOOLDIR + 20];
    long		uniqueval;	/* Unique value to return */

    (void) sprintf  (uniquepath, "%s/%s", SPOOLDIR, UNIQUEFILE);
    (void) sprintf  (lockpath, "%s/%s", SPOOLDIR, ULOCKFILE);
    if ((ufd = open (uniquepath, 2)) < 0  &&  errno == ENOENT) {
	if ((ufd = creat (uniquepath, 0644)) < 0) {
	    (void) fprintf (stderr, "Couldn't creat %s\n", uniquepath);
	    return 0;
	}
	uniqueval = 0;
	(void) write (ufd, (char *) &uniqueval, sizeof uniqueval);
	(void) lseek (ufd, 0L, 0);
    }
    lockit (uniquepath, lockpath);
    if (read (ufd, (char *) &uniqueval, sizeof uniqueval) != sizeof uniqueval) {
	(void) fprintf (stderr, "%s was wrong size\n", uniquepath);
	uniqueval = 0;
    }
    uniqueval++;
    (void) lseek (ufd, 0L, 0);
    (void) write (ufd, (char *) &uniqueval, sizeof uniqueval);
    (void) close (ufd);
    unlockit (uniquepath, lockpath);
    return uniqueval;
}


/*
 * The following code was cribbed directly from Geoff Kuenning's fifo
 * stuff for the cron daemon.
 */
#define MAXSLEEP	63	/* Maximum sleep is 63 seconds */

static void lockit (path, lock)	/* Lock 'path' using links to 'lock' */
char *path;			/* Path name to be locked */
char *lock;			/* Lock file to create */
{
    static unsigned long ranseed;	/* Seed for 'random' numbers */
    static int		ranset = 0;	/* NZ if ranseed has been set */
    unsigned int	sleepmask = 1;	/* Exponentially growing sleep time */

    while (link (path, lock) < 0  &&  errno == EEXIST) {
	if (!ranset) {
	    ranset = 1;
	    ranseed = time ((long *) 0);
	}
	ranseed = ranseed * 1001 + 123;	/* Multiplicative congruential */
	/* Delay a random time */
	(void) sleep (sleepmask & (unsigned int) (ranseed >> 8));
	if (sleepmask < MAXSLEEP) /* Increase sleep time */
	    sleepmask |= sleepmask << 1;
    }
}

/* ARGSUSED */
static void unlockit (path, lock) /* Unlock 'path' using 'lock' */
char * path;			/* Path name to be unlocked */
char * lock;			/* Path name of lock file */
{
    (void) unlink (lock);	/* That wasn't hard, was it? */
}
#endif
