#ifndef lint
static char Rcs_Id[] =
	"$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/batch/batchd.c,v 1.9 1999/07/14 05:22:31 johnh Exp $";
#endif

/*
 *  Batch daemon
 */

/*
 * $Log: batchd.c,v $
 * Revision 1.9  1999/07/14 05:22:31  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 1.8  1995/11/27 05:03:56  geoff
 * Properly declare a bunch of static routines.  Support HASNLIST_UNION
 *
 * Revision 1.7  87/03/31  23:13:25  geoff
 * Add code to calculate load averages the hard way on USG systems that
 * have the sysinfo data structure.
 * 
 * Revision 1.6  87/01/10  23:23:29  geoff
 * More changes to improve portability.
 * 
 * Revision 1.5  86/11/29  19:16:20  geoff
 * Add (untested) code to support systems that do not have SIGCLD or SIGCHLD.
 * 
 * Revision 1.4  86/11/29  15:58:06  geoff
 * Major changes to make it more configureable.  Mostly this consists of
 * replacing BSD ifdef's with more feature-specific ones.
 * 
 * Revision 1.3  86/07/06  23:59:10  geoff
 * Fix a whole bunch of lint complaints.
 * 
 * Revision 1.2  86/06/26  23:29:48  geoff
 * Major changes to make it portable to System V.  Also add some hysteresis
 * in the load scheduling by using a shorter-term load average for stopping
 * than for starting.
 * 
 */

#define GEN_ID

#include "config.h"
#include "lex.h"
#include <errno.h>
#include <ctype.h>
#include <signal.h>
#include <pwd.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#ifndef PARAMGETSTYPES
#include <sys/types.h>
#endif
#ifdef NULL
#undef NULL
#endif
#include <stdio.h>
#include <sys/file.h>
#include <sys/stat.h>
#ifdef HASNLIST_H
#include <nlist.h>
#else
#include <a.out.h>
#endif
#ifdef HASRLIMIT
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#endif
#include "ndir.h"
#ifdef HASFCNTL
#include <fcntl.h>
#endif
#ifdef HASSYSINFO
#include <sys/sysinfo.h>
#include <sys/times.h>
#endif

#ifdef RELIABLE_SIGNALS
/*
 * For 4.2bsd signals.
 */
#define	mask(s)		(1 << ((s)-1))
#define	sighold(s)	sigblock(mask(s))
#define	sigrelse(s)	sigsetmask(sigblock(0)&~mask(s))
#endif

#undef DIRSIZ
#define	DIRSIZ	32	/* Used to be 14; could be MAXNAMLEN. -IAN! */

char Pid[] = PIDFILE;
#ifdef DEBUG
#define STATIC
#else
#define STATIC	static
#endif

STATIC SIG_TYPE alarming ();
#ifdef SIGCHLD
STATIC SIG_TYPE reapchild ();
#else
STATIC int reapchild ();
#endif
STATIC pxerror ();
STATIC pmerror ();
STATIC FILE *sendmail ();
STATIC mailclose ();
STATIC readqueues ();
STATIC readpro ();
STATIC requeue ();
STATIC runqueue ();
STATIC startjob ();
STATIC rltoi ();
STATIC Ktorl ();
STATIC itorl ();
STATIC terminated ();
STATIC mailback ();
STATIC stopall ();
STATIC startall ();
STATIC abortall ();
STATIC abortjob ();
STATIC rundown ();

#ifndef R_OK
#define R_OK	04		/* To test if read access is OK */
#endif

STATIC int nkids;

STATIC struct stat dot, odot;	/* Start of spooldir */
STATIC int alarmed;

typedef char MAIL;
#define	MAIL_START	1
#define	MAIL_END	2
#define	MAIL_CRASH	4

typedef struct {
	char	f_name[DIRSIZ+1];
	struct stat f_stat;
	int	f_pid;
	char	f_seen;
} CF_File;

#define NRLIM 6		/* Number of resources, e.g. RLIMIT_CPU */

typedef struct {
	char	q_name[DIRSIZ+1];
	CF_File q_files[MAXENTRIES];/* Sorted list of files */
	int	q_running;	/* If not set, queue is not active */
	int	q_idle;		/* If set, queue doesn't have anything in it */
	int	q_drain;	/* Set if shutdown started */
	int	q_abort;	/* Set if waiting for aborting */
#ifdef HASLOADAVG
#ifdef SIGSTOP
	int	q_stopped;	/* Set if queue stopped for lower load */
#endif
	int	q_overload;	/* Set if high load stopped scheduling */
#endif
	struct	stat q_stat, q_ostat;	/* Stat of queue directory */
	struct	stat q_prostat, q_oprostat; /* Stat of queue profile file */
	char	*q_supervisor;	/* Queue supervisor */
	MAIL	q_supmail;	/* Type of mail for supervisor */
	MAIL	q_usermail;	/* Type of notification for user */
	int	q_maxexec;	/* Max # of executing jobs */
	int	q_nexec;	/* Current # of executing jobs */
#ifdef HASLOADAVG
	double	q_loadhalt;	/* Stop scheduling new jobs if over this load */
	double	q_loadsched;	/* Restart scheduling jobs if over this load */
	int	q_lhaltind;	/* Index of load to use for loadhalt test */
	int	q_lschind;	/* Index of load to use for loadsched test */
#ifdef SIGSTOP
	/*
	 * The following two names are not unique within 8 characters.
	 * Fortunately, they are only used on BSD, where the compiler can
	 * handle it.  (Thank you, Berkeley).
	 */
	double	q_loadstop;	/* Stop jobs if over this load */
	double	q_loadstart;	/* Stop jobs if over this load */
	int	q_lstopind;	/* Index of load to use for loadstop test */
	int	q_lstartind;	/* Index of load to use for loadstart test */
#endif
#endif
	int	q_nice;		/* Job priority */
	int	q_restart;	/* Restart on system crash? */
	int	q_pgroup;	/* Process group for this queue */
#ifdef HASRLIMIT
	struct rlimit q_rlimit[NRLIM];	/* Resource hard/soft limits */
#endif
#ifdef ULIMIT
	long	q_flimit;		/* File size limit */
#endif
#ifdef NOT_IMPLEMENTED
	/* Remainder not yet implemented */
	struct {
		time_t	q_starttime;	/* Start queue at */
		time_t	q_endtime;	/* End queue at */
	} q_times[5];
	char	q_days[7];	/* Day to activate on, sun = 0 */
#endif NOT_IMPLEMENTED
} Queue;
Queue queues[NQUEUES];

typedef struct {
	int pid;
	Queue *rqp;
	CF_File *rfp;
} Running;
Running running[MAXTOTAL];

STATIC int ignore[] = { SIGINT, SIGHUP, SIGQUIT,
#ifndef linux
			SIGSYS,
#endif
			SIGPIPE, SIGTERM };

extern unsigned int alarm ();
extern char *calloc ();
extern int errno;
extern void exit ();
extern char *getenv ();
extern struct passwd *getpwnam ();
extern char *strchr ();
extern long lseek ();
extern void perror ();
extern char *strrchr ();
extern char *strcpy ();
extern char *strncpy ();
extern daddr_t ulimit ();

#ifdef HASLOADAVG
STATIC void calcloads ();
STATIC double loads[3];
extern double exp ();
#endif
STATIC FILE *mfatal ();
STATIC FILE *muser ();

main()
{
	reg1 Queue *qp;
	char fname[DIRSIZ*2];
	reg2 int i;
	int dotu;
	int pidno;
#ifdef HASSYSINFO
	long now;
	long then;
#endif
	FILE *f;
	FILE *mail;
	extern char **environ;
	static char *env[] = { "PATH=/bin", 0, 0 };

#ifndef DEBUG
	if(getuid() != 0) {
		mail = mfatal ();
		(void) fprintf (mail, "Fatal error; batch terminating\n");
		(void) fprintf (mail, "Real uid is %d, not root\n", getuid ());
		fexit (mail);
	}
	if(geteuid() != 0) {
		mail = sendmail (SUPERVISOR);
		(void) fprintf (mail, "Fatal error; batch terminating\n");
		(void) fprintf (mail,
		  "Effective uid is %d, not root\n", geteuid ());
		mailclose (mail);
		exit (1);
	}
#endif
	if(chdir(SPOOLDIR) < 0)
		pxerror(SPOOLDIR);
	dotu = open(".", O_RDONLY, 0);
	if (dotu < 0)
		pxerror("open .");
	if(fstat(dotu, &dot) < 0)
		pxerror("stat of dot");
#ifndef DEBUG
	if(dot.st_uid != 0) {
		mail = sendmail (SUPERVISOR);
		(void) fprintf (mail, "Fatal error; batch terminating\n");
		(void) fprintf (mail, "%s: Not owned by root\n", SPOOLDIR);
		mailclose (mail);
		exit (1);
	}
#endif
	if(dot.st_mode & (S_IWRITE>>6)) {
		mail = sendmail (SUPERVISOR);
		(void) fprintf (mail, "Fatal error; batch terminating\n");
		(void) fprintf (mail, "%s: Mode %03o allows general write\n",
		    SPOOLDIR, dot.st_mode);
		mailclose (mail);
#ifdef DEBUG
		(void) abort ();
#else
		exit (1);
#endif
	}
#ifndef DEBUG
	
	/* Go into background */
	if (fork ())
	    exit (0);
	/* Log my pid */
	f = fopen (Pid, "r");
	if (f != NULL) {
		if (fscanf (f, "%d", &pidno) == 1) {
			if (kill (pidno, 0) == 0) {
				mail = sendmail (SUPERVISOR);
				(void) fprintf (mail,
				    "Fatal error; batch terminating\n");
				(void) fprintf (mail,
				    "%s exists.  Batchd probably already running. Please check.\n",
				    Pid) ;
				mailclose (mail);
				exit (1);
			}
		}
		(void) fclose (f);
	}
	if (f = fopen(Pid, "w")) {
		(void) fprintf (f, "%d\n", getpid ());
		(void) fclose (f);
	}
	/*
	 * Turn off signals, and set up an appropriate nice level.
	 */
	for (i = 0; i < sizeof ignore/sizeof ignore[0]; i++)
		(void) signal (ignore[i], SIG_IGN);
#ifdef RELNICE
	(void) nice (-(MAXNICE + 1 - MINNICE));	/* Go to minimum nice level */
#endif
	(void) nice ((MAXNICE + 1 - MINNICE) / 2); /* Pick an in-between nice */
	/*
	 *  Get rid of controlling tty
	 */
#ifdef TIOCNOTTY
	if((i = open("/dev/tty", 2)) != -1) {
		(void) ioctl (i, TIOCNOTTY, 0);
		(void) close (i);
	}
#endif
#ifdef HASSETPGRP
#ifdef SETPGRPARGS
	(void) setpgrp (0, getpid ());
#else
	(void) setpgrp ();
#endif
#endif
#endif

	/*
	 *  Get rid of most current environment variables.
	 */
	if (getenv ("TZ"))
		env[1] = getenv ("TZ") - 3;
	environ = env;

	(void) signal (SIGALRM, alarming);
#ifdef SIGCHLD
	(void) signal (SIGCHLD, reapchild);
#endif
	for(;;) {
		/*
		 *  Check if the main spool directory changed.  If so, make
		 *  a new list of potential queue entries.
		 */
		if(fstat(dotu, &dot) < 0)
			pxerror("statting dot");
		if(dot.st_mtime > odot.st_mtime) {
			readqueues();
			odot = dot;
		}
		/*
		 *  Now check each potential queue entry.
		 */
		for(qp = queues ; qp < &queues[NQUEUES] ; qp++) {
			if(qp->q_name[0] == 0)
				continue;
			/*
			 *  Check if its profile file changed; if so re-read it.
			 *  Be generous, if somebody deleted it don't affect
			 *  the queue or whats runing.
			 */
			(void) sprintf (fname, "%s/%s", qp->q_name, PROFILE);
			if(stat(fname, &qp->q_prostat) < 0) {
				pmerror(fname);
				continue;
			}
			if(qp->q_prostat.st_mtime > qp->q_oprostat.st_mtime) {
				readpro(qp);
				qp->q_oprostat = qp->q_prostat;
			}
			/*
			 *  If profile file said exec off, ignore.
			 *  If still draining then we have to call runqueue
			 *  still for purposes of starting and stopping
			 */
			if(qp->q_drain) {
				runqueue(qp);
				continue;
			}
			if(qp->q_running == 0)
				continue;
			/*
			 *  Now check the directory itself.  If it changed,
			 *  then something new might have been queued.
			 */
			if(stat(qp->q_name, &qp->q_stat) < 0) {
				pmerror(fname);
				continue;
			}
			if(qp->q_stat.st_mtime > qp->q_ostat.st_mtime) {
				requeue(qp);
				qp->q_ostat = qp->q_stat;
			}
			/*
			 *  Now, start something running in queue if it can.
			 */
			runqueue(qp);
		}
		/*
		 *  Now wait for children until either there are none; or
		 *  NSEC seconds have gone by.
		 */
#ifdef HASSYSINFO
		then = time ((long *) NULL);
		(void) alarm (MINLOADTIME);
#else
		(void) alarm (NSEC);
#endif
		alarmed = 0;
		while (!alarmed) {
#ifdef SIGCHLD
			(void) pause ();
#else
			if (reapchild () < 0  &&  errno == ECHILD)
				(void) pause ();
#endif
#ifdef HASSYSINFO
			if (alarmed) {
				calcloads ();
				now = time ((long *) NULL);
				if (now - then < NSEC) {
					alarmed = 0;
					(void) alarm (MINLOADTIME);
				}
			}
#endif
		}
		(void) alarm (0);
#ifdef DEBUG
		mail = muser (DEBUGFILE);
		(void) fprintf (mail, "looking around.\n");
		mailclose (mail);
#endif DEBUG
	}
}

#ifdef SIGCHLD
STATIC void
reapchild()
#else
STATIC int
reapchild()
#endif
{
#ifdef DEBUG
	FILE *mail;
#endif
	int pid;
#ifdef HASRLIMIT
	union wait status;

	while ((pid = wait3(&status, WNOHANG, (struct rusage *)0)) > 0) {
#ifdef DEBUG
		mail = muser (DEBUGFILE);
		(void) fprintf (mail, "wait3 exit pid=%d, stat=%o\n",
			pid, status.w_status);
		mailclose (mail);
#endif DEBUG
		terminated(pid, status.w_status);
	}
#else /* ~HASRLIMIT */
	int status;

	while ((pid = wait (&status)) > 0) {
#ifdef DEBUG
		mail = muser (DEBUGFILE);
		(void) fprintf (mail, "wait exit pid=%d, stat=%o\n",
			pid, status);
		mailclose (mail);
#endif DEBUG
		terminated(pid, status);
	}
#endif
#ifdef SIGCLD
	/*
	 * On BSD systems, the signal is named SIGCHLD and is not reset
	 * when caught.  On USG, it's SIGCLD, and it is reset when caught,
	 * so we have to set it again.
	 */
	(void) signal (SIGCLD, reapchild);
#endif
#ifndef SIGCHLD
	return pid;		/* Caller must be able to detect wait failure */
#endif
}

STATIC SIG_TYPE
alarming()
{

	++alarmed;
	(void) signal (SIGALRM, alarming);
}

#ifdef linux
#include <errno.h>
#else /* ! linux */
extern int	sys_nerr;
extern char	*sys_errlist[];
#endif /* linux */

STATIC
pxerror(s)
char *s;
{
	reg1 char *c;
	reg2 FILE *mail;

	c = "Unknown error";
	if(errno < sys_nerr)
		c = sys_errlist[errno];
	mail = mfatal ();
	(void) fprintf (mail, "%s: %s\n", s, c);
	fexit (mail);
}

STATIC
pmerror(s)
char *s;
{
	reg1 char *c;
	reg2 FILE *mail;
	int err;

	err = errno;
	mail = sendmail(SUPERVISOR);
	c = "Unknown error";
	if(err < sys_nerr)
		c = sys_errlist[err];
	(void) fprintf (mail, "%s: %s\n", s, c);
	mailclose(mail);
}

STATIC FILE *
mfatal ()
{
	FILE *mail;

	mail = sendmail (SUPERVISOR);
	(void) fprintf (mail, "Fatal error; batch terminating\n");
	return mail;
}

fexit (mail)
FILE *mail;
{
	mailclose (mail);
#ifdef DEBUG
	abort ();
#else
	exit (1);
#endif

}

STATIC FILE *
muser (username)
char *username;
{
	FILE *mail;
	time_t now, time();
	char *ctime();

	if (username == NULL)
		return NULL;
	if (*username == '/') {
		if ((mail = fopen (username, "a")) == NULL) {
			pmerror (username);
			mail = stderr;
		}
		now = time(0);
		(void) fprintf (mail, "%-15.15s ", ctime (&now) + 4);
	} else
		mail = sendmail (username);
	return mail;
}

STATIC FILE *
sendmail(s)
char *s;
{
	reg2 FILE *mailer;
	int pid;
	struct {int read; int write;} mailp;
	reg1 Running *rp;

	if(pipe(&mailp) < 0) {
	err:
		(void) fprintf (stderr, "%s: ", s);
		return stderr;
	}
	switch(pid = vfork()) {
	case 0:
#ifdef HASDUP2
		(void) dup2 (mailp.read, fileno(stdin));
#else
		(void) close (fileno(stdin));
		(void) dup (mailp.read);
#endif
		(void) close (mailp.write);
		(void) nice (4);	/* lower priority */
		(void) execl (MAILER, MAILERNAME, s, (char *)0);
		perror("BATCHD execl MAILER!");
		exit(1);
	case -1:
		goto err;
	default:
		(void) close (mailp.read);
	}
	nkids++;
	for(rp = running ; rp < &running[MAXTOTAL] ; rp++)
		if(rp->pid == 0) {
			rp->pid = pid;
			rp->rqp = NULL;
			rp->rfp = NULL;
			break;
		}
	if(rp >= &running[MAXTOTAL])
		(void) fprintf (stderr, "BATCHD MAILER: too many jobs\n");
	mailer = fdopen(mailp.write, "w");
	if(mailer == NULL) {
		(void) close (mailp.write);
		perror("BATCHD fdopen");
		goto err;
	}
	(void) fprintf (mailer, "To: %s\nFrom: batchd (Batch Daemon)\n\n", s);
	return mailer;
}

STATIC
mailclose(f)
FILE *f;
{
	(void) fclose (f);
}

/*
 *  Read the home directory, looking for all current queue entries.
 */
STATIC
readqueues()
{
	reg1 Queue *qp;
	reg2 Queue *nextq;
	reg3 int i;
	static DIR *dotp;
	struct direct *d;
	char seen[NQUEUES];
	struct stat sb;
	reg4 CF_File *fp;

	if(dotp == NULL) {
		if((dotp = opendir(".")) == NULL)
			pxerror("opendir dot");
	} else
		rewinddir(dotp);
	(void) memset (seen, 0, sizeof seen);
	while((d = readdir(dotp)) != NULL) {
		/*
		 *  Ignore everything starting with a dot.
		 */
		if(*d->d_name == '.')
			continue;
		if(stat(d->d_name, &sb) == -1) {
			pmerror(d->d_name);
			continue;
		}
		if((sb.st_mode & S_IFMT) != S_IFDIR) {
			errno = ENOTDIR;
			pmerror(d->d_name);
			continue;
		}
		/*
		 *  Found a directory.
		 */
		nextq = NULL;
		for(qp = queues ; qp < &queues[NQUEUES] ; qp++) {
			if(strcmp(d->d_name, qp->q_name) == 0)
				break;
			if(nextq == NULL  &&  qp->q_name[0] == 0)
				nextq = qp;
		}
		/*
		 *  Was it a new queue?
		 */
		if(qp >= &queues[NQUEUES]) {
			if(nextq == NULL) {
				FILE *mail;

				mail = sendmail(SUPERVISOR);
				(void) fprintf (mail,
				  "More than %d queues at %s\n",
				  NQUEUES, d->d_name);
				mailclose(mail);
				continue;
			}
			qp = nextq;
			(void) strncpy (qp->q_name, d->d_name, DIRSIZ);
			qp->q_nexec = 0;
#ifdef HASLOADAVG
#ifdef SIGSTOP
			qp->q_stopped = 0;
#endif
			qp->q_overload = 0;
#endif
			qp->q_ostat.st_mtime = qp->q_oprostat.st_mtime = 0;
			for(fp = qp->q_files;
			    fp < &qp->q_files[MAXENTRIES];
			    fp++)
				fp->f_name[0] = 0;
		}
		seen[qp-queues] = 1;
	}
	/*
	 *  Any directories which have been deleted imply the queues should
	 *  end violently since we can't get the output files..
	 */
	for(i = 0 ; i < NQUEUES ; i++)
		if(seen[i] == 0  &&  queues[i].q_name[0]) {
			if(qp->q_nexec) {
				abortall(&queues[i]);
				qp->q_abort = 1;
			} else {
				qp->q_name[0] = 0;
				qp->q_pgroup = 0;
			}
		}
}

/*
 *  Read a profile file under the given queue entry.
 */
STATIC
readpro(qp)
reg1 Queue *qp;
{
#ifdef HASRLIMIT
	static struct rlimit minus_one = { -1, -1 };
	reg4 int i;
#endif
	int run;
	FILE *pro, *mail;
	char pname[DIRSIZ*2];
	reg2 enum keyword k;
	extern char yytext[];
	extern double atof ();
	MAIL *mailp;

	/*
	 *  First set up queue to all defaults so if we re-read a file
	 *  we don't get the old values.
	 */
	qp->q_supervisor = NULL;
	qp->q_usermail = qp->q_supmail = MAIL_END|MAIL_CRASH;
	qp->q_nice = 0;
	qp->q_restart = 0;
#ifdef HASRLIMIT
	for( i=0; i<NRLIM; i++ )
		qp->q_rlimit[i] = minus_one;
#endif
#ifdef ULIMIT
	qp->q_flimit = 0x7FFFFFFFL;
#endif
#ifdef NOT_IMPLEMENTED
	qp->q_times[0].q_starttime = 0;
	qp->q_days[0] = qp->q_days[1] = qp->q_days[2] = qp->q_days[3] =
		qp->q_days[4] = qp->q_days[5] = qp->q_days[6] = 0;
#endif NOT_IMPLEMENTED
	qp->q_drain = qp->q_abort = 0;
	qp->q_maxexec = 1;
	qp->q_idle = 0;
#ifdef HASLOADAVG
	qp->q_loadhalt = DEFHALT;
	qp->q_loadsched = DEFSCHED;
	qp->q_lhaltind = DEFHALTAVERAGE;
	qp->q_lschind = DEFSCHEDAVERAGE;
#ifdef SIGSTOP
	qp->q_loadstop = DEFSTOP;
	qp->q_loadstart = DEFSTART;
	qp->q_lstopind = DEFSTOPAVERAGE;
	qp->q_lstartind = DEFSTARTAVERAGE;
#endif
#endif
	if(qp->q_pgroup == 0) {
		reg3 Queue *qp1;
		static int nextpg = 31100;	/* Random number higher than
						   system allocates for procs*/
		again:
		for(qp1 = queues ; qp1 < &queues[NQUEUES] ; qp1++)
			if(qp1->q_pgroup == nextpg) {
				nextpg++;
				goto again;
			}
		qp->q_pgroup = nextpg++;
		if(nextpg > 31200)
			nextpg = 31100;
	}
	run = 1;	/* Default is start running */

	/*
	 *  Try to read the profile file.  If not there, assume shutdown this
	 *  queue.
	 */
	(void) sprintf (pname, "%s/%s", qp->q_name, PROFILE);
	if((pro = fopen(pname, "r")) == NULL) {
		pxerror(pname);
		rundown(qp);
		return;
	}
	lexfile(pro);
#define	syntax	goto syntaxerr
#define	lex	(k = (enum keyword) yylex ())
	while((int)lex) {
#ifdef HASRLIMIT
		reg5 int rli;
		reg6 int rl;
		reg7 int limit;
#endif
		switch(k) {
		case K_LINE:
			break;
		case K_EXEC:
			switch(lex) {
			case K_OFF:
				run = 0;
				break;
			case K_LINE:
			case K_ON:
				run = 1;
				break;
			case K_DRAIN:
				run = 0;
				qp->q_drain = 1;
				break;
			default:
				syntax;
			}
			break;
		case K_MAXEXEC:
			if(lex != K_NUMBER)
				syntax;
			qp->q_maxexec = atoi(yytext);
			break;
		case K_SUPERVISOR:
			if(lex != K_VARIABLE)
				syntax;
			qp->q_supervisor =
			  calloc (1, (unsigned) strlen (yytext) + 1);
			(void) strcpy (qp->q_supervisor, yytext);
			break;
		case K_MAIL:
			mailp = &qp->q_usermail;
		mailstuff:
			*mailp = 0;
			if(lex == K_LINE)
				break;
			if(k != K_VARIABLE)
				syntax;
			if(strchr(yytext, 's'))
				*mailp |= MAIL_START;
			if(strchr(yytext, 'e'))
				*mailp |= MAIL_END;
			if(strchr(yytext, 'c'))
				*mailp |= MAIL_CRASH;
			break;
		case K_MAILSUPERVISOR:
			mailp = &qp->q_supmail;
			goto mailstuff;
#ifdef HASLOADAVG
		case K_LOADSCHED:
			if(lex != K_NUMBER)
				syntax;
			qp->q_loadhalt = qp->q_loadsched = atof (yytext);
			qp->q_lhaltind = DEFHALTAVERAGE;
			qp->q_lschind = DEFSCHEDAVERAGE;
			if (lex == K_LINE)
				break;
			else if (k != K_NUMBER)
				syntax;
			qp->q_loadsched = atof (yytext);
			if (lex == K_LINE)
				break;
			else if (k != K_NUMBER)
				syntax;
			qp->q_lhaltind = atoi (yytext);
			if (lex == K_LINE)
				break;
			else if (k != K_NUMBER)
				syntax;
			qp->q_lschind = atoi (yytext);
			break;
#ifdef SIGSTOP
		case K_LOADSTOP:
			if(lex != K_NUMBER)
				syntax;
			qp->q_loadstart = qp->q_loadstop = atof (yytext);
			qp->q_lstopind = DEFSTOPAVERAGE;
			qp->q_lstartind = DEFSTARTAVERAGE;
			if (lex == K_LINE)
				break;
			else if (k != K_NUMBER)
				syntax;
			qp->q_lstartind = atof (yytext);
			if (lex == K_LINE)
				break;
			else if (k != K_NUMBER)
				syntax;
			qp->q_lstopind = atoi (yytext);
			if (lex == K_LINE)
				break;
			else if (k != K_NUMBER)
				syntax;
			qp->q_lstartind = atoi (yytext);
			break;
#endif
#endif
		case K_NICE:
			if(lex != K_NUMBER)
				syntax;
			qp->q_nice = atoi(yytext);
			break;
		case K_RESTART:
			qp->q_restart = 1;
			break;
#ifdef HASRLIMIT
		case K_CPULIMIT:
		case K_RLIMCPU:
		case K_RLIMDATA:
		case K_RLIMSTACK:
		case K_RLIMCORE:
		case K_RLIMRSS:
#ifndef ULIMIT
		case K_RLIMFSIZE:
#endif /* ULIMIT */
			/* If only one number is given, we set only the
			 * current limit.  If two numbers, we set current and
			 * maximum limits.   -IAN!
			 */
			rli = rltoi (rl = Ktorl (k)); /* idx in q_rlimit arr */
			if(lex != K_NUMBER)
				syntax;
			limit = atoi(yytext);
			if(lex == K_LINE){
				if (getrlimit (rl, &(qp->q_rlimit[rli])) != 0) {
					pmerror("getrlimit(%d,%d)\n",
					    rl, &(qp->q_rlimit[rli]));
					break;
				}
			}
			qp->q_rlimit[rli].rlim_cur = limit;
			if(k == K_LINE)
				break;
			if(k != K_NUMBER)
				syntax;
			qp->q_rlimit[rli].rlim_max = atoi(yytext);
			break;
#endif
#ifdef ULIMIT
		case K_RLIMFSIZE:
			if(lex != K_NUMBER)
				syntax;
			qp->q_flimit = atoi (yytext);
			break;
#endif
		default:
			syntax;
		}
		if(k != K_LINE)
			if ((enum keyword) yylex () != K_LINE)
				syntax;
	}
	(void) fclose (pro);
	if(run == 0)
		if(qp->q_drain)
			rundown(qp);
		else
			abortall(qp);
	qp->q_running = run;
	return;

syntaxerr:
	mail = sendmail (SUPERVISOR);
	(void) fprintf (mail,
	  "%s: Syntax error near %s\n", qp->q_name, yytext);
	mailclose (mail);
	rundown(qp);
}

/*
 *  Read the directory for a queue and find all jobs ready to run.
 */
STATIC
requeue(qp)
reg1 Queue *qp;
{
	reg5 DIR *qdir;
	reg4 struct direct *d;
	reg6 FILE *mail;
	char procname[50];
	reg2 CF_File *fp;
	reg3 CF_File *nextf;
	static int startup = 1;

	for(fp = qp->q_files ; fp < &qp->q_files[MAXENTRIES] ; fp++)
		fp->f_seen = 0;
	if((qdir = opendir(qp->q_name)) == NULL) {
		pmerror(qp->q_name);
		rundown(qp);
		return;
	}
	while((d = readdir(qdir)) != NULL) {
		if(strncmp(d->d_name, "cf", 2) != 0)
			continue;
		nextf = NULL;
		for(fp = qp->q_files ; fp < &qp->q_files[MAXENTRIES] ; fp++) {
			if(fp->f_name[0] == 0) {
				/* Found the first empty slot */
				if(nextf == NULL)
					nextf = fp;
				continue;
			}
			if(strcmp(d->d_name, fp->f_name) == 0)
				break;
		}
		/*
		 * If we ran off the end of the table, we didn't
		 * find it and it must be a new entry.
		 */
		if(fp >= &qp->q_files[MAXENTRIES]) {
			if(nextf == NULL) {
				mail = muser (qp->q_supervisor);
				(void) fprintf (mail,
				   "%s: More than %d entries; %s ignored\n",
				   qp->q_name, MAXENTRIES, d->d_name);
				mailclose (mail);
				continue;
			}
			(void) sprintf (procname,
			  "%s/%s", qp->q_name, d->d_name);
			fp = nextf;
			if(stat(procname, &fp->f_stat) == -1) {
				pmerror(procname);
				continue;
			}
			(void) strncpy (fp->f_name, d->d_name, DIRSIZ);
			fp->f_pid = 0;
		}
		fp->f_seen = 1;
	}
	/*
	 * Look at our internal table and check that every job we
	 * know about is still in the directory.  If one has been
	 * deleted and is currently running, abort it; otherwise
	 * just remove the job from the table.
	 */
	for(fp = qp->q_files ; fp < &qp->q_files[MAXENTRIES] ; fp++){
		if(fp->f_seen == 0  &&  fp->f_name[0])
			if(fp->f_pid)
				abortjob(fp);
			else
				fp->f_name[0] = 0;
	}
	/*
	 * Check for restarted jobs.  This is done exactly once, when
	 * BATCHD gets started after a reboot.
	 */
	if(startup) {
		startup = 0;
		rewinddir(qdir);
		while((d = readdir(qdir)) != NULL) {
			if(strncmp(d->d_name, "of", 2) != 0)
				continue;
			for(fp = qp->q_files ; fp < &qp->q_files[MAXENTRIES] ; fp++)
				if(strcmp(d->d_name+1, fp->f_name+1) == 0) {
					char cfname[DIRSIZ*2];
					mailback(0, qp, fp, MAIL_CRASH,
					qp->q_restart?"restarted":"not restarted");
					if(qp->q_restart)
						break;
					/*
					 *  If restart disallowed, delete file
					 */
					(void) sprintf (cfname, "%s/%s",
					    qp->q_name, fp->f_name);
					if(unlink(cfname) == -1) {
						int e = errno;

						pmerror(cfname);
						mail =
						    muser (qp->q_supervisor);
						(void) fprintf (mail,
						  "%s: Can't unlink dead job%s",
						    qp->q_name,
							e==ENOENT? "":
							"; stopping queue");
						mailclose (mail);
						if (e != ENOENT)
							rundown(qp);
					}
					fp->f_name[0] = 0;
					break;
				}
			/*
			 *  Left over output files with no control files??
			 */
			if(fp >= &qp->q_files[MAXENTRIES]) {
				char ofname[DIRSIZ*2];
				(void) sprintf (ofname,
				  "%s/%s", qp->q_name, d->d_name);
				(void) unlink (ofname);
				mail = muser (qp->q_supervisor);
				(void) fprintf (mail,
				    "%s: Old output file deleted\n", ofname);
				mailclose (mail);
			}
		}
	}
	(void) closedir (qdir);
	qp->q_idle = 0;
}

/*
 *  A queue that might have changed.
 *  Wander along it and start jobs if possible.
 *  If we couldn't fork we just return; in 30 seconds we'll try again.
 */
STATIC
runqueue(qp)
reg1 Queue *qp;
{
	reg2 CF_File *fp;
	reg3 CF_File *bestfp;

#ifdef HASLOADAVG
#ifdef SIGSTOP
	calcloads ();
	if(qp->q_stopped  &&  loads[qp->q_lstartind] < qp->q_loadstart
	    &&  loads[qp->q_lstopind] < qp->q_loadstop) {
		startall(qp);
		qp->q_stopped = 0;
	}
	if(qp->q_stopped == 0  &&  loads[qp->q_lstopind] >= qp->q_loadstop) {
		stopall(qp);
		qp->q_stopped = 1;
	}
	if(qp->q_stopped)
		return;
#endif
#endif
	if(qp->q_drain  ||  qp->q_running == 0  ||  qp->q_idle)
		return;
#ifdef HASLOADAVG
	if(qp->q_overload  &&  loads[qp->q_lschind] < qp->q_loadsched)
		qp->q_overload = 0;
	if(qp->q_overload == 0  &&  loads[qp->q_lhaltind] >= qp->q_loadhalt)
		qp->q_overload = 1;
	if (qp->q_overload)
		return;
#endif
	qp->q_idle = 1;
	while(qp->q_nexec < qp->q_maxexec) {
		bestfp = NULL;
		for(fp = qp->q_files ; fp < &qp->q_files[MAXENTRIES] ; fp++)
			if(fp->f_name[0]  &&  fp->f_pid == 0) {
				if(bestfp == NULL ||
				   strcmp(bestfp->f_name+2, fp->f_name+2) > 0)
					bestfp = fp;
			}
		if(bestfp != NULL) {
			if(startjob(qp, bestfp) == 0) {
				qp->q_idle = 0;
				return;
			}
		} else
			break;
	}
	qp->q_idle = 1;
	return;
}

/*
 *  Start an actual job executing.
 */
STATIC
startjob(qp, fp)
reg1 CF_File *fp;
reg2 Queue *qp;
{
	int pid;
	reg4 int i;
	FILE *mail;
	char outname[50];
	char procname[50];
	int fd;
	reg3 Running *rp;
	struct passwd *pw, *getpwuid();

	(void) sprintf (procname, "%s/%s", qp->q_name, fp->f_name);
	if ((pw = getpwuid ((int) fp->f_stat.st_uid)) == NULL)
		return 1;	/* ? */
	fp->f_name[0] = 'o';	/* Output to `of' file name */
	(void) sprintf (outname, "%s/%s", qp->q_name, fp->f_name);
	fp->f_name[0] = 'c';
#ifdef RELIABLE_SIGNALS
	sighold(SIGCHLD);		/* until in parent table */
#endif
	if((pid = vfork()) == -1) {
#ifdef RELIABLE_SIGNALS
		sigrelse(SIGCHLD);
#endif
		pmerror("forking");
		return 0;
	}
	if(pid) {
#ifdef DEBUG
		mail = muser (DEBUGFILE);
		(void) fprintf (mail, "Job %s, pid %d\n", procname, pid);
		mailclose (mail);
#endif DEBUG
		nkids++;
		qp->q_nexec++;
		fp->f_pid = pid;
		for(rp = running ; rp < &running[MAXTOTAL] ; rp++)
			if(rp->pid == 0) {
				rp->pid = pid;
				rp->rqp = qp;
				rp->rfp = fp;
				break;
			}
#ifdef RELIABLE_SIGNALS
		sigrelse(SIGCHLD);	/* job is safely in table now */
#endif
		if(rp >= &running[MAXTOTAL]) {
			mail = sendmail(SUPERVISOR);
			(void) fprintf (mail, "More than %d jobs; too many\n",
			    MAXTOTAL);
			mailclose(mail);
		}
		if (qp->q_usermail & MAIL_START) {
			mail = muser (pw->pw_name);
			(void) fprintf (mail,
			  "%s: Job is starting now.\n", procname);
			mailclose (mail);
		}
		if (qp->q_supmail & MAIL_START) {
			mail = muser (qp->q_supervisor);
			(void) fprintf (mail, "%s: Job starting for %s\n",
			    procname, pw->pw_name);
			mailclose (mail);
		}
		return 1;
	}
#ifdef RELIABLE_SIGNALS
	sigrelse(SIGCHLD);	/* for symmetry and completeness */
#endif
	/*
	 *  Child. Setuid to the owner and setup
	 *  i/o redirection stuff to mail it back.
	 */
#ifdef HASSETPGRP
#ifdef SETPGRPARGS
	(void) setpgrp (0, qp->q_pgroup);
#else
	(void) setpgrp ();
#endif
#endif
	if ((fd = creat(outname, 0600)) < 0) {
#ifndef vfork
		/*
		 * On vfork systems, we share data space with our parent, and
		 * can't do stdio.  So if we have a problem, we must fork
		 * again to report it.  All this because Berkeley couldn't
		 * do copy-on-write.
		 */
		fd = errno;		/* Preserve errno past fork */
		if (fork () > 0)	/* Ignore errors */
		    exit(1);
		errno = fd;
#endif
		pxerror(outname);
	}
#ifdef HASDUP2
	(void) dup2 (fd, 1);
	(void) dup2 (fd, 2);
#else
	(void) close (1);
	(void) close (2);
	(void) dup (fd);
	(void) dup (fd);
#endif
	for(i=3 ; i < NOFILE ; i++)
		(void) close (i);
	(void) close (0);
	if (open ("/dev/null", 0) != 0) {
		(void) close (0);
		(void) open ("/", 0);
	}
#ifdef HASRLIMIT
	for( i=0; i<NRLIM; i++ ){
		reg5 struct rlimit *rlp = &(qp->q_rlimit[i]);
		if( rlp->rlim_cur >= 0 && rlp->rlim_max >= 0 )
			setrlimit( itorl(i), rlp );
	}
#endif
#ifdef ULIMIT
	(void) ulimit (1, qp->q_flimit);
#endif
	(void) nice (qp->q_nice);
#ifdef INITGROUPS
	initgroups(pw->pw_name, fp->f_stat.st_gid);
#endif
	(void) setgid ((int) fp->f_stat.st_gid);
	(void) setuid ((int) fp->f_stat.st_uid);
	for(i=1 ; i <= NSIG ; i++)
		(void) signal (i, SIG_DFL);
	(void) execlp (procname, procname, (char *) 0);
#ifndef vfork
	/*
	 * On vfork systems, we must fork before calling pxerror.
	 * What a pain.
	 */
	fd = errno;		/* Preserve errno past fork */
	if (fork () > 0)	/* Ignore errors */
	    exit(1);
	errno = fd;
#endif
	pxerror(procname);
	/*NOTREACHED*/
}

#ifdef HASRLIMIT
/* Turn RLIMIT manifest number into Integer.
 */
STATIC int
rltoi( rl )
int rl;
{
	FILE *mail;

	switch( rl ){
	case RLIMIT_CPU: return 0;
	case RLIMIT_FSIZE: return 1;
	case RLIMIT_DATA: return 2;
	case RLIMIT_STACK: return 3;
	case RLIMIT_CORE: return 4;
	case RLIMIT_RSS: return 5;
	}
	mail = mfatal ();
	(void) fprintf (mail, "rltoi(%d) invalid\n", rl );
	fexit (mail);
	/*NOTREACHED*/
}
#endif

#ifdef HASRLIMIT
/* Turn K token from LEX into RLIMIT number.
 */
STATIC int
Ktorl( k )
enum keyword k;
{
	FILE *mail;

	switch( k ){
	case K_CPULIMIT: return(RLIMIT_CPU);
	case K_RLIMCPU: return(RLIMIT_CPU);
	case K_RLIMFSIZE: return(RLIMIT_FSIZE);
	case K_RLIMDATA: return(RLIMIT_DATA);
	case K_RLIMSTACK: return(RLIMIT_STACK);
	case K_RLIMCORE: return(RLIMIT_CORE);
	case K_RLIMRSS: return(RLIMIT_RSS);
	}
	mail = mfatal ();
	(void) fprintf (mail, "Ktorl(%d) invalid\n", k );
	fexit (mail);
	/*NOTREACHED*/
}
#endif

#ifdef HASRLIMIT
/* Turn Integer into RLIMIT number.
 */
STATIC int
itorl( i )
int i;
{
	FILE *mail;

	switch( i ){
	case 0: return(RLIMIT_CPU);
	case 1: return(RLIMIT_FSIZE);
	case 2: return(RLIMIT_DATA);
	case 3: return(RLIMIT_STACK);
	case 4: return(RLIMIT_CORE);
	case 5: return(RLIMIT_RSS);
	}
	mail = mfatal ();
	(void) fprintf (mail, "itorl(%d) invalid\n", i);
	fexit (mail);
	/*NOTREACHED*/
}
#endif

/*
 *  A process terminated.
 */
STATIC
terminated(pid, status)
{
	reg1 Running *rp;
	reg2 Queue *qp;
	reg3 CF_File *fp;
	char cfname[DIRSIZ*2];
	reg4 FILE *mail;

	for(rp = running ; rp < &running[MAXTOTAL] ; rp++)
		if(rp->pid == pid)
			break;
	if(rp >= &running[MAXTOTAL]) {
		/* This should never happen, provided SIGCHLD is
		 * properly held from before the fork() until after
		 * the process pid is put in the run table.   -IAN!
		 */
		mail = sendmail(SUPERVISOR);
		(void) fprintf (mail,
		    "Signal %d from process %d but process not in internal table\n",
			status, pid);
		mailclose(mail);
		return;
	}
	nkids--;
	qp = rp->rqp;
	/*  Assume this is return of a mail process */
	if(qp == NULL) {
		rp->pid = 0;
		return;
	}
	fp = rp->rfp;
	mailback(status, qp, fp, MAIL_END, (char *)0);
	if(qp->q_supmail & MAIL_END) {
		mail = muser (qp->q_supervisor);
		(void) fprintf (mail, "%s/%s: Job ended, status %o\n",
		    qp->q_name, fp->f_name, status);
		mailclose (mail);
	}
	(void) sprintf (cfname, "%s/%s", qp->q_name, fp->f_name);
	if(unlink(cfname) == -1) {
		int e = errno;
		
		pmerror(cfname);
		mail = muser (qp->q_supervisor);
		(void) fprintf (mail, "%s: Can't unlink jobs%s\n", qp->q_name,
			e==ENOENT? "": "; stopping queue");
		mailclose (mail);
		if (e != ENOENT)
			rundown(qp);
	}
	fp->f_name[0] = 0;
	fp->f_pid = 0;
	rp->pid = 0;
	qp->q_nexec--;
	if(qp->q_nexec < 0) {
		mail = mfatal ();
		(void) fprintf (mail, "nexec < 0!\n");
		fexit (mail);
	}
	/*
	 *  Finally all gone??
	 */
	if(qp->q_abort)
		if(qp->q_nexec == 0) {
			qp->q_name[0] = 0;
			qp->q_pgroup = 0;
		}
	/*
	 *  Drained?
	 */
	if(qp->q_drain  &&  qp->q_nexec == 0) {
		qp->q_drain = 0;
		mail = muser (qp->q_supervisor);
		(void) fprintf (mail, "%s: Drained finally\n", qp->q_name);
		mailclose (mail);
	}
	qp->q_idle = 0;
	runqueue(qp);
}

STATIC
mailback(status, qp, fp, mailstat, expl)
Queue *qp;
CF_File *fp;
MAIL mailstat;
char *expl;
{
	FILE *mail;
	char buf[BUFSIZ];
	char outname[DIRSIZ*2];
	int n;
	int output;
	int outstat;
	struct stat sb;
	struct passwd *pw, *getpwuid();
	/*
	 * Signames.h contains a string table named "faults" which, when
	 * indexed by signal number, will describe that signal.  The
	 * header file is generated by signames.SH.  If you get syntax
	 * errors, feel free to edit it as appropriate to cure them.
	 */
#include "signames.h"

	fp->f_name[0] = 'o';
	(void) sprintf (outname, "%s/%s", qp->q_name, fp->f_name);
	fp->f_name[0] = 'c';
	outstat = stat(outname, &sb);
	/*
	 *  Send mail even if we aren't supposed to if there is output
	 *  or status is non-null.
	 */
	if((qp->q_usermail & mailstat) == 0  &&
	   (outstat==-1 || sb.st_size==0)  &&
	   status == 0) {
		if(unlink(outname) < 0)
			pmerror(outname);
		return;
	}
	if ((pw = getpwuid ((int) fp->f_stat.st_uid)) == NULL) {
		mail = sendmail(SUPERVISOR);
		(void) fprintf (mail, "Can't mail %s to %d\n",
		    outname, fp->f_stat.st_uid);
		mailclose(mail);
		return;
	}
	/*
	 * If submitter is root, mail to $USER.
	 * This ain't pretty...
	 */
	if (pw->pw_uid == 0) {
		FILE *f;
		char *strrchr();

		strrchr(outname, '/')[1] = 'c';
		if ((f = fopen(outname, "r")) != NULL) {
			char *p, *up = NULL;
			struct passwd *xpw, savepw;
			
			while (fgets(buf, sizeof buf, f) != NULL) {
				if (strncmp(buf, "USER=", 5) == 0) {
					up = buf+5;
					break;
				}
				if (strncmp(buf, "setenv USER ", 12) == 0) {
					up = buf+12;
					break;
				}
			}
			(void) fclose (f);
			if (up == NULL)
				goto nope;
			/* crack 'user' */
			while (*up && *up != '\'')
				up++;
			if (*up == 0)
				goto nope;
			p = ++up;
			while (*p && *p != '\'')
				p++;
			if (*p == 0)
				goto nope;
			*p = 0;
			savepw = *pw;
			xpw = getpwnam(up);
			if (xpw != NULL)
				pw = xpw;
			else
				*pw = savepw;
		}
	   nope:
		strrchr(outname, '/')[1] = 'o';
	}
	mail = sendmail(pw->pw_name);
	(void) fprintf (mail,
	  "Job %s in %s queue has ", fp->f_name, qp->q_name);
	if(mailstat & MAIL_CRASH)
		(void) fprintf (mail,
		  "been aborted due to crash; it was %s", expl);
	else {
	    (void) fprintf (mail, "completed");
	    if(status == 0)
		    (void) fprintf (mail, " successfully");
	    else {
		    if(status & 0177) {
			    if((status & 0177) > NSIG)
				    (void) fprintf (mail,
				      " with unknown status %o", status);
			    else
				    (void) fprintf (mail,
				      " with termination %s",
				      faults[(status & 0177)]);
			    if(status & 0200)
				    (void) fprintf (mail, " and a core dump");
		    } else
			    (void) fprintf (mail,
			      " with exit code %d", status >> 8);
	    }
	}
	(void) fprintf (mail, ".\n\n");
	if(outstat == -1)
		(void) fprintf (mail, "Can't stat output file %s\n", outname);
	else if(sb.st_size > 0) {
		(void) fprintf (mail, "Output %sfollows:\n\n",
		    mailstat & MAIL_CRASH ? "to crash point " : "");
		if((output = open(outname, 0)) == -1) {
			pmerror(outname);
			(void) fprintf (mail,
			  "**Daemon can't read the output data!\n");
		} else {
			while((n = read(output, buf, sizeof(buf))) > 0)
				(void) fwrite (buf, n, 1, mail);
			(void) close (output);
		}
	}
	if(unlink(outname) < 0)
		pmerror(outname);
	mailclose(mail);
}

#ifdef HASLOADAVG
#ifdef SIGSTOP
/*
 *  Stop all jobs in a particular queue.
 */
STATIC
stopall(qp)
reg1 Queue *qp;
{
	reg2 FILE *mail;

	if(qp->q_nexec == 0)
		return;
	mail = muser (qp->q_supervisor);
	(void) fprintf (mail, "%s: Stopping; load too high\n", qp->q_name);
	mailclose (mail);
	killpg(qp->q_pgroup, SIGSTOP);
}
#endif

#ifdef SIGSTOP
/*
 *  Restart all jobs in a particular queue.
 */
STATIC
startall(qp)
reg1 Queue *qp;
{
	reg2 FILE *mail;

	if(qp->q_nexec == 0)
		return;
	mail = muser (qp->q_supervisor);
	(void) fprintf (mail, "%s: Restarting; load is ok now\n", qp->q_name);
	mailclose (mail);
	killpg(qp->q_pgroup, SIGCONT);
}
#endif
#endif
	
/*
 *  Abort all jobs in a particular queue.
 */
STATIC
abortall(qp)
reg1 Queue *qp;
{
	reg2 CF_File *fp;
	reg3 FILE *mail;

	if(qp->q_nexec == 0)
		return;
	mail = muser (qp->q_supervisor);
	(void) fprintf (mail,
	  "%s:  Terminating all jobs with extreme prejudice", qp->q_name);
	mailclose (mail);
	for(fp = qp->q_files ; fp < &qp->q_files[MAXENTRIES] ; fp++)
		if(fp->f_pid)
			abortjob(fp);
}

/*
 *  Abort a particular job.  Use extreme prejudice.
 *  Note that we count on wait getting the process back later.
 */
STATIC
abortjob(fp)
reg1 CF_File *fp;
{
	(void) kill (fp->f_pid, SIGKILL);
}

/*
 *  Queue is to be terminated. Stop new jobs running in that queue.
 */
STATIC
rundown(qp)
reg1 Queue *qp;
{
	reg2 FILE *mail;

	if(qp->q_running) {
		mail = muser (qp->q_supervisor);
		(void) fprintf (mail, "%s: Draining\n", qp->q_name);
		mailclose (mail);
	}
	qp->q_running = 0;
	if(qp->q_nexec == 0) {
		mail = muser (qp->q_supervisor);
		(void) fprintf (mail, "%s: Drained\n", qp->q_name);
		mailclose (mail);
		qp->q_drain = 0;
	}
}

#ifdef HASLOADAVG
#ifdef HASNLIST_UNION
struct	nlist nl[2];
#else
struct	nlist nl[] = {
	{ AVENRUN },
	{ 0 },
};
#endif
#define	X_AVENRUN	0

STATIC void
calcloads()
{
	static int kmem = -1;
#ifdef HASSYSINFO
	static NRUNTYPE avenrun[3];
	double avg;
	double decay;
	struct tms dummy;
	struct sysinfo info;
	static long lastrunque;
	static long lastrunocc;
	static time_t lasttime = 0;
	time_t now;
#else
	NRUNTYPE avenrun[3];
#endif

	if(kmem == -1) {
#ifdef HASNLIST_UNION
		nl[0].n_un.n_name = AVENRUN;
		nl[1].n_un.n_name = NULL;
#endif
		(void) nlist (NAMELIST, nl);
		if (nl[0].n_type==0) {
			FILE *mail;

			mail = sendmail(SUPERVISOR);
			(void) fprintf (mail, "%s: No namelist\n", NAMELIST);
			mailclose(mail);
			return;
		}
		if((kmem = open("/dev/kmem", 0)) == -1) {
			pmerror("/dev/kmem");
			return;
		}
	}
	(void) lseek (kmem, (long) nl[X_AVENRUN].n_value, 0);
#ifdef HASSYSINFO
	(void) read (kmem, (char *) &info, sizeof info);
	/*
	 * We must calculate the load average ourselves, based on the
	 * data kept in 'struct sysinfo'.  System V provides several
	 * possibilities for a load average:  process switches per
	 * second, swapin/out activity, various sorts of I/O activity,
	 * system calls, fork/exec, and run/swap queue sizes.  BSD
	 * uses the average run queue length;  we will use the same.
	 * As in BSD, we will exponentially decay the result.
	 */
	now = times (&dummy);	/* This gets the system tick counter */
	if (now - lasttime < (MINLOADTIME - 1) * TICKS)
		return;		/* Don't sample ridiculously often */
	if (lasttime == 0) {
		avenrun[0] = avenrun[1] = avenrun[2] =
		    (double) info.runque / info.runocc;
	}
	else {
		lastrunocc = info.runocc - lastrunocc;
		if (lastrunocc <= 0)
			return;			/* Should never happen */

#if 1		/* I decided higher loads are more conservative */
		lastrunque = info.runque - lastrunque;
#else
		/* By subtracting lastrunocc, we exclude ourselves. */
		lastrunque = info.runque - lastrunque - lastrunocc;
#endif
		if (lastrunque < 0)
			lastrunque = 0;
		avg = (double) lastrunque / lastrunocc;
		/*
		 * Formula for digital simulation of exponential decay:
		 *
		 * <new value> = <old value> * p + <sample value> * (1 - p)
		 *
		 * where p is calculated such that a value of 1 will decay
		 * to a value of 1/e in TC seconds.  ('e' is the base of
		 * natural logarithms;  TC is the "time constant" for the
		 * decay:  in this case, 1, 5, or 15 minutes).  It is thus
		 * easy to see that the formula p = exp (-t/TC) will produce
		 * exp (-1) == e when t == TC.
		 */
		lasttime -= now;	/* This makes a negative no. */
		decay = exp ((double) lasttime / (TC0 * TICKS));
		avenrun[0] = avenrun[0] * decay + avg * (1 - decay);
		decay = exp ((double) lasttime / (TC1 * TICKS));
		avenrun[1] = avenrun[1] * decay + avg * (1 - decay);
		decay = exp ((double) lasttime / (TC2 * TICKS));
		avenrun[2] = avenrun[2] * decay + avg * (1 - decay);
	}
#if 0
	{
	long clock;
	time (&clock);
	(void) fprintf (stderr, "load average: %.2f %.2f %.2f at %s",
	avenrun[0], avenrun[1], avenrun[2], ctime (&clock));
	}
#endif
	lastrunque = info.runque;
	lastrunocc = info.runocc;
	lasttime = now;
#else
	(void) read (kmem, (char *) avenrun, sizeof (avenrun));
#endif
	loads[0] = (double) avenrun[0] / NRUNFACTOR;
	loads[1] = (double) avenrun[2] / NRUNFACTOR;
	loads[2] = (double) avenrun[2] / NRUNFACTOR;
}
#endif
