#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/collect.c,v 2.13 1999/07/27 01:22:00 johnh Exp $";
#endif

/*
 * $Log: collect.c,v $
 * Revision 2.13  1999/07/27 01:22:00  johnh
 * location support for at jobs
 *
 * Revision 2.12  1995/11/27 01:59:23  geoff
 * Include errno.h for the debugging macros.
 *
 * Revision 2.11  1995/05/11  05:05:03  geoff
 * Minor RCS header changes.
 *
 * Revision 2.10  87/01/10  19:13:45  geoff
 * Fix a lint complaint.
 * 
 * Revision 2.9  86/11/29  19:50:13  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.8  86/03/30  15:40:48  geoff
 * Switch debugging output to the DBPRINTn macros when possible
 * 
 * Revision 2.7  86/03/30  14:32:44  geoff
 * Take out a lot of unnecessary re-collect requests.  Also fix a bug
 * (ce_input should never be passed as an argument to free).
 * 
 * Revision 2.6  86/03/28  14:40:34  geoff
 * Add support for UF_NEVERDEL.
 * 
 * Revision 2.5  86/03/25  19:55:15  geoff
 * Free ce_command only if it exists, now that it's optional.
 * 
 * Revision 2.4  86/01/04  16:15:33  geoff
 * Add collection of the pending-command queues.
 * 
 * Revision 2.3  85/11/25  16:50:54  geoff
 * Add a missing declaration.
 * 
 * Revision 2.2  85/11/24  16:11:41  geoff
 * Expand the debug output.
 * 
 * Revision 2.1  85/11/04  23:39:31  geoff
 * Initial revision.
 * 
 */

#include <errno.h>
#include <stdio.h>
#include "cron.h"

void		collect ();		/* Perform garbage collection */
static void	mark ();		/* Mark stuff for deletion */
static void	keep ();		/* Unmark stuff we need to keep */
static void	zapents ();		/* Delete cronents & unmark files */
static void	zapfiles ();		/* Delete files & unmark users */
static void	zapusers ();		/* Delete users */

short		re_collect = 0;		/* NZ if collection left some stuff */

#ifdef DEBUG
extern short	debug;			/* Debug control */
extern void	dumpall ();		/* Dump all data structures */
#endif
extern struct cronent
		enthead;		/* Head of cronent chain */
extern void	endpwent ();		/* Finish looking at passwd file */
extern struct cronfile
		filehead;		/* Head of cronfile chain */
extern void	free ();		/* Return memory to the allocator */
extern struct childproc
		prochead;		/* Head of child process list */
extern struct cronuser
		userhead;		/* Head of user chain */

/*
 * Garbage collector for cron.  This routine searches through cron's
 * data structures looking for ones that can be deallocated, and freeing
 * those that can.
 *
 * cronent - free when DELETED and childproc doesn't point to it
 * cronfile - free when DELETED and cronent doesn't point to it
 * prochead - free when child process dies
 * cronuser - free when no file points to it
 *
 * First, we set the MARKDEL bits in all the struct cronusers and in those
 * struct cronents and struct cronfiles that have the EF_DELETED or FF_DELETED
 * bit set.  Then we search the process list, clearing MARKDEL bits
 * for cronents pointed to by active processes.  Next, we search
 * the cronent list, freeing anything marked for delete and clearing
 * the MARKDEL bits for files pointed to by the remaining entries.
 * Following this with the file list, we delete marked files and mark
 * the users owning the other files.  Finally, we search the user
 * list and delete them.
 */
void collect ()				/* Garbage collector */
    {

    DBPRINT0 (4, "collect ()\n");
    re_collect = 0;			/* Record that we have done coll. */
    mark ();				/* Mark stuff for deletion */
    keep ();				/* Unmark stuff we need to keep */
    zapents ();				/* Delete cronents & unmark files */
    zapfiles ();			/* Delete files & unmark users */
    zapusers ();			/* Delete users */
#ifdef DEBUG
    if (debug >= 14)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "Results after collection:\n");
	dumpall ();
	}
#endif
    endpwent ();			/* Make sure password file is closed */
    }

static void mark ()			/* Mark stuff for deletion */
    {
    reg1 struct cronent *	entp;	/* Crontab line being marked */
    reg2 struct cronfile *	filep;	/* Cron file being marked */
    reg3 struct cronuser *	userp;	/* Cron user being marked */

    for (entp = enthead.ce_next;  entp != &enthead;  entp = entp->ce_next)
	{
	if (entp->ce_flags & EF_DELETED)
	    entp->ce_flags |= EF_MARKDEL;
	}
    for (filep = filehead.cf_next;
      filep != &filehead;
      filep = filep->cf_next)
	{
	if (filep->cf_flags & FF_DELETED)
	    filep->cf_flags |= FF_MARKDEL;
	}
    for (userp = userhead.cu_next;
      userp != &userhead;
      userp = userp->cu_next)
	{
	if ((userp->cu_flags & UF_NEVERDEL) == 0)
	    userp->cu_flags |= UF_MARKDEL;
	}
    }

static void keep ()			/* Unmark stuff we need to keep */
    {
    reg2 struct cmdqueue *	cmdp;	/* Current command queue entry */
    reg3 struct childproc *	procp;	/* Current entry on child list */
    reg1 struct cronuser *	userp;	/* Current user entry */

    /*
     * Search outstanding processes and preserve their crontab entries
     * against their termination.
     */
    for (procp = prochead.cp_next;
      procp != &prochead;
      procp = procp->cp_next)
	procp->cp_ent->ce_flags &= ~EF_MARKDEL;
    /*
     * Search pending commands and preserve the entries so that the commands
     * can be run when the system load drops.
     */
    for (userp = userhead.cu_next;
      userp != &userhead;
      userp = userp->cu_next)
	{
	for (cmdp = userp->cu_queue.cq_next;
	  cmdp != &(userp->cu_queue);
	  cmdp = cmdp->cq_next)
	    cmdp->cq_ent->ce_flags &= ~EF_MARKDEL;
	}
    }

static void zapents ()			/* Delete cronents & unmark files */
    {
    reg1 struct cronent *	entp;	/* Current crontab entry */
    reg2 struct cronent *	nentp;	/* Next crontab entry, for deleting */

    for (entp = enthead.ce_next;  entp != &enthead;  entp = nentp)
	{
	nentp = entp->ce_next;
	if (entp->ce_flags & EF_MARKDEL) /* Is this one gone for good? */
	    {
	    entp->ce_prev->ce_next = nentp; /* Yes, delete it */
	    nentp->ce_prev = entp->ce_prev;
	    if (entp->ce_command)
		free (entp->ce_command);
#ifdef LOC_CRON
	    if (entp->ce_location)
		free (entp->ce_location);
#endif /* LOC_CRON */
	    free (entp);
	    }
	else				/* No, preserve its file */
	    entp->ce_file->cf_flags &= ~FF_MARKDEL;
	}
    }

static void zapfiles ()			/* Delete files & unmark users */
    {
    reg1 struct cronfile *	filep;	/* Current file entry */
    reg2 struct cronfile *	nfilep;	/* Next file entry, for deleting */

    for (filep = filehead.cf_next;  filep != &filehead;  filep = nfilep)
	{
	nfilep = filep->cf_next;
	if (filep->cf_flags & FF_MARKDEL) /* Is this one gone for good? */
	    {
	    filep->cf_prev->cf_next = nfilep; /* Yes, delete it */
	    nfilep->cf_prev = filep->cf_prev;
	    free (filep->cf_name);
	    free (filep);
	    }
	else				/* No, preserve its user */
	    filep->cf_user->cu_flags &= ~UF_MARKDEL;
	}
    }

static void zapusers ()			/* Delete users */
    {
    reg1 struct cronuser *	userp;	/* Current user entry */
    reg2 struct cronuser *	nuserp;	/* Next user entry, for deleting */

    for (userp = userhead.cu_next;  userp != &userhead;  userp = nuserp)
	{
	nuserp = userp->cu_next;
	if (userp->cu_flags & UF_MARKDEL) /* Is this one gone for good? */
	    {
	    userp->cu_prev->cu_next = nuserp; /* Yes, delete it */
	    nuserp->cu_prev = userp->cu_prev;
	    free (userp->cu_home);
	    free (userp->cu_name);
	    free (userp);
	    }
	}
    }
