#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/crontab.c,v 2.15 1998/07/30 20:42:23 geoff Exp $";
#endif

/*
 * $Log: crontab.c,v $
 * Revision 2.15  1998/07/30 20:42:23  geoff
 * Many changes, accumulated over far too much time: add security code to
 * try to avoid unlink/rename race attacks, and add long-name support.
 *
 * Revision 2.14  1995/11/27  01:59:34  geoff
 * Add the -e switch.
 *
 * Revision 2.13  1995/05/11  05:05:17  geoff
 * Minor RCS header changes.
 *
 * Revision 2.12  87/04/01  23:50:29  geoff
 * Tell cron to read the file whenever it is listed, to help protect
 * naive users from themselves.
 * 
 * Revision 2.11  87/01/10  19:14:44  geoff
 * Don't complain if cron isn't running when it doesn't matter.  Create the
 * crontab in a temporary file, so even "cat /usr/spool/cron/crontabs/xxx |
 * crontab -u xxx" will work without destroying the crontab.
 * 
 * Revision 2.10  86/12/01  14:45:24  geoff
 * Add a missing declaration (thank you, lint!).
 * 
 * Revision 2.9  86/11/29  19:50:27  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.8  86/11/16  00:58:04  geoff
 * Fix a minor lint complaint.
 * 
 * Revision 2.7  86/04/17  21:55:20  geoff
 * Add the -L switch and allow/deny support.
 * 
 * Revision 2.6  86/03/28  14:41:19  geoff
 * Fix a minor bug that made crontab refuse argument-less invocations.
 * 
 * Revision 2.5  86/03/25  19:55:49  geoff
 * Tear out a whole bunch of stuff that's been moved to sendcmd.c so that
 * 'at' can use it too.  Also add support for crontabs on standard input.
 * 
 * Revision 2.4  86/01/04  16:16:18  geoff
 * Add the -u switch.  Straighten out some chdir's.
 * 
 * Revision 2.3  85/12/21  15:15:04  geoff
 * Add the setlibdir routine and calls to it.  Improve the documentation for
 * sendcommand.
 * 
 * Revision 2.2  85/11/25  16:51:31  geoff
 * Fix some lint complaints.
 * 
 * Revision 2.1  85/11/24  15:44:04  geoff
 * Initial revision.
 * 
 */

#define GEN_ID				/* Generate source-control ID's */

#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include "cron.h"

int		main ();		/* "Crontab" command */
static void	ctusage ();		/* Issue usage message */
static int	listtab ();		/* List a cron table */
static int	removetab ();		/* Remove a cron table */
static int	edittab ();		/* Edit a cron table */
static int	addtab ();		/* Add a cron table */
static int	showlog ();		/* Print cron log */

char *		progname;		/* Name we are running under */

extern char *	errmsg ();		/* Return system error message */
extern char *	getenv ();		/* Get environment variable */
extern char *	getname ();		/* Get user's name */
extern char *	optarg;			/* Option argument pointer */
extern int	optind;			/* Index of argument to process */
extern int	sendcommand ();		/* Send command to the cron daemon */
extern void	setlibdir ();		/* Change directories to LIBDIR */
extern char *	strchr ();		/* Search for character in string */
extern struct passwd * userpwent;	/* User's password entry */
					/* Set by getname */

int main (argc, argv)			/* Crontab command */
    reg2 int		argc;		/* Argument count */
    reg1 char *		argv[];		/* Argument vector */
    {
    int			edit = 0;	/* NZ to edit our crontab */
    int			list = 0;	/* NZ to list our crontab */
    reg3 int		nextopt;	/* Next option from getopt */
    int			printlog = 0;	/* NZ to print from log file */
    int			remove = 0;	/* NZ to remove our crontab */
    reg4 char *		user = NULL;	/* User's name */

    progname = argv[0];
    while ((nextopt = getopt (argc, argv, "erLlu:")) != EOF)
	{
	switch (nextopt)
	    {
	    case 'e':			/* Edit crontab */
		edit = 1;
		break;
	    case 'r':			/* Remove crontab */
		remove = 1;
		break;
	    case 'L':			/* Print log file */
		printlog = 1;
		break;
	    case 'l':			/* List crontab */
		list = 1;
		break;
	    case 'u':
		if (getuid ())
		    {
		    (void) fprintf (stderr,
		      "-u option may only be used by root\n");
		    return 1;
		    }
		user = optarg;		/* Explicit user name/ID */
		break;
	    default:
		ctusage ();
		return 1;
	    }
	}
    if (edit + list + remove + printlog > 1)
	{
	ctusage ();
	return 1;
	}
    if ((edit | list | remove | printlog)  &&  optind < argc)
	{
	ctusage ();
	return 1;
	}

    if (!(edit | list | remove | printlog)  &&  optind < argc - 1)
	{
	ctusage ();
	return 1;
	}

    user = getname (user);		/* Get user's name */
    if (user == NULL)
	return 1;

    if (!allowed (user, CRONPRIV))
	{
	(void) fprintf (stderr,
	  "Sorry, you are not allowed to submit cron tables\n");
	return 2;
	}

    if (edit)
	return edittab (user);
    else if (list)
	return listtab (user);
    else if (remove)
	return removetab (user);
    else if (printlog)
	return showlog (user);
    else
	return addtab (user, optind < argc ? argv[optind] : "-");
    }

static void ctusage ()		/* Put out a usage message */
    {
    (void) fprintf (stderr,
      "Usage:\tcrontab -l [-u user]\n\tcrontab -r [-u user]\n");
    (void) fprintf (stderr, "\tcrontab -e [-u user]\n");
    (void) fprintf (stderr, "\tcrontab -L [-u user]\n\n");
    (void) fprintf (stderr, "\tcrontab [-u user] file\n");
    }

static int listtab (user)	/* List a cron table */
    char *		user;	/* Name of owner of the table */
    {
    reg1 FILE *		cfile;	/* Crontab file */
    reg2 int		ch;	/* Character read from file */

    setlibdir ();
    (void) chdir (CRONTABDIR);
    cfile = fopen (user, "r");
    (void) chdir (LIBDIR);
    if (cfile == NULL)
	{
	(void) fprintf (stderr, "Couldn't open cron table for %s\n", user);
	return 1;
	}
    while ((ch = getc (cfile)) != EOF)
	(void) putchar (ch);
    (void) fclose (cfile);
    /*
     * Now tell the cron daemon to read the file.  This shouldn't be
     * necessary, but it turns out that people assume that if "crontab -l"
     * shows the table they want, the cron daemon knows about it.  Since
     * this is a perfectly reasonable assumption, the easy solution is to
     * make sure that it's true by re-adding the table whenever it's
     * listed.  This adds a bit to overhead, but you can't have everything.
     * It is also less than perfect, because the user probably typed
     * "crontab -l" after his command failed to run, but again you can't
     * have everything.  (It may turn out that because of the latter,
     * this is a misfeature.  We'll find out.)
     */
    if (sendcommand ('a', user, 1) == SEND_ERROR)
	return 1;
    else
	return 0;
    }

static int removetab (user)	/* Remove a cron table */
    char *		user;	/* Name of owner of the table */
    {
    int			retstat; /* Status to return to caller */

    setlibdir ();
    retstat = 0;
    if (sendcommand ('d', user, 0) == SEND_ERROR)
	retstat = 1;
    (void) chdir (CRONTABDIR);
    retstat |= unlink (user);
    (void) chdir (LIBDIR);
    return retstat;
    }

static int edittab (user)		/* Edit a cron table */
    char *		user;		/* Name of owner of the table */
    {
    reg3 int		ch;		/* Character being copied */
    reg1 FILE *		cfile;		/* Access to original crontab file */
    int			donepid;	/* Process that exited */
    char		editcommand[1024]; /* Command to use to edit */
    char *		editor;		/* Editor to execute */
    int			kidpid;		/* Child process' PID */
    reg2 FILE *		outfile;	/* Pointer for writing cron copy */
    int			status;		/* Statis of child process */
    char		tempname[sizeof EDITTEMP]; /* Space for tmp name */
    struct stat		tempstat1;	/* Stat buffer for temp file */
    struct stat		tempstat2;	/* Stat bfr for temp file afterwards */

    setlibdir ();
    (void) chdir (CRONTABDIR);
    cfile = fopen (user, "r");
    if (cfile == NULL  &&  errno != ENOENT)
	{
	(void) fprintf (stderr, "Couldn't open cron table for %s: %s\n", user,
	  errmsg (0));
	return 1;
	}
    (void) strcpy (tempname, EDITTEMP);	/* Set up a temp file name */
    (void) tempfile (tempname);		/* Create the temp file name */
    if ((outfile = fopen (tempname, "w")) == NULL)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Can't create temp file '%s' to hold crontab: %s\n",
	  tempname, errmsg (0));
	if (cfile != NULL)
	    (void) fclose (cfile);
	return 1;
	}
    if (fstat (fileno(outfile), &tempstat1) < 0)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Can't stat temp file '%s': %s\n", tempname, errmsg (0));
	if (cfile != NULL)
	    (void) fclose (cfile);
	(void) fclose (outfile);
	(void) unlink (tempname);
	return 1;
	}
    (void) chmod (tempname, 0600);
    (void) chown (tempname, userpwent->pw_uid, userpwent->pw_gid);
    /*
     * Protect against an unlink/rename attack by ensuring the i-node
     * number hasn't changed.
     */
    if (stat (tempname, &tempstat2) < 0)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Can't stat temp file '%s': %s\n", tempname, errmsg (0));
	if (cfile != NULL)
	    (void) fclose (cfile);
	(void) fclose (outfile);
	(void) unlink (tempname);
	return 1;
	}
    if (tempstat1.st_dev != tempstat2.st_dev
      ||  tempstat1.st_ino != tempstat2.st_ino
      ||  tempstat1.st_nlink != tempstat2.st_nlink)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Temp file '%s' has stat mismatch, suspect security attack\n",
	  tempname);
	if (cfile != NULL)
	    (void) fclose (cfile);
	(void) fclose (outfile);
	/*
	 * We don't unlink tempname, to prevent being tricked into
	 * unlinking something important.
	 */
	return 1;
	}
    if (fstat (fileno(outfile), &tempstat1) < 0)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Can't stat temp file '%s': %s\n", tempname, errmsg (0));
	if (cfile != NULL)
	    (void) fclose (cfile);
	(void) fclose (outfile);
	(void) unlink (tempname);
	return 1;
	}
    if (tempstat1.st_uid != userpwent->pw_uid
      ||  tempstat1.st_gid != userpwent->pw_gid)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Temp file '%s' has UID/GID mismatch, suspect security attack\n",
	  tempname);
	if (cfile != NULL)
	    (void) fclose (cfile);
	(void) fclose (outfile);
	/*
	 * We don't unlink tempname, to prevent being tricked into
	 * unlinking something important.
	 */
	return 1;
	}
    if (cfile != NULL)
	{
	while ((ch = getc (cfile)) != EOF)
	    (void) putc (ch, outfile);
	(void) fclose (cfile);
	}
    (void) fclose (outfile);
    /*
     * Spawn an editor for the user.  We have to create the subprocess
     * ourselves, because we're currently set-uid, and we want to run
     * the subprocess in the user's home directory if possible.  But
     * after that we'll use system() to spawn the editor.  This is
     * superior because it allows the EDITOR environment variable to
     * be a complete shell command.
     */
    kidpid = fork ();
    if (kidpid == -1)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr, "Can't fork: %s\n", errmsg (0));
	(void) unlink (tempname);
	return 1;
	}
    else if (kidpid != 0)
	{
	/*
	 * Parent.  Just wait for the child process.  Fail if it does.
	 */
	while ((donepid = wait (&status)) != kidpid)
	    {
	    if (donepid == -1)
		{
		(void) chdir (LIBDIR);
		(void) fprintf (stderr,
		  "Editor process didn't complete: %s\n", errmsg (0));
		(void) unlink (tempname);
		return 1;
		}
	    }
	if (status != 0)
	    {
	    (void) chdir (LIBDIR);
	    (void) fprintf (stderr, "Editor failed: %d\n", status);
	    (void) unlink (tempname);
	    return 1;
	    }
	}
    else
	{
	/*
	 * Child.  Set up a safe environment, and then run the user's
	 * preferred editor.
	 */
	setgid (userpwent->pw_gid);
	setuid (userpwent->pw_uid);
	if (userpwent->pw_dir == NULL  ||  chdir (userpwent->pw_dir) < 0)
	    chdir (NOHOME);
	editor = getenv ("VISUAL");
	if (editor == NULL)
	    editor = getenv ("EDITOR");
	if (editor == NULL)
	    editor = DEFEDITOR;
	(void) strncpy (editcommand, editor, sizeof editcommand - 1);
	(void) strncat (editcommand, " ",
	  sizeof editcommand - strlen (editcommand) - 1);
	(void) strncat (editcommand, tempname,
	  sizeof editcommand - strlen (editcommand) - 1);
	editcommand[sizeof editcommand - 1] = '\0';
	status = system (editcommand);
	exit (status);
	}
    /*
     * We get here only in the parent, and only if the editor
     * succeeded.  We can now just run addtab to insert the new table.
     * But first we must protect against a possible substitution of
     * the temp file.
     */
    if (stat (tempname, &tempstat2) < 0)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Can't stat temp file '%s': %s\n", tempname, errmsg (0));
	(void) unlink (tempname);
	return 1;
	}
    if (tempstat1.st_dev != tempstat2.st_dev
      ||  tempstat1.st_ino != tempstat2.st_ino
      ||  tempstat1.st_nlink != tempstat2.st_nlink
      ||  tempstat1.st_uid != tempstat2.st_uid
      ||  tempstat1.st_gid != tempstat2.st_gid)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Temp file '%s' has stat mismatch, possible security attack.\n",
	  tempname);
	(void) fprintf (stderr, "Continue anyway? [n]");
	ch = getchar();
	if (ch != 'y'  &  ch != 'Y')
	    return 1;
	}
    status = addtab (user, tempname);
    if (stat (tempname, &tempstat2) < 0)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Can't stat temp file '%s': %s\n", tempname, errmsg (0));
	(void) fprintf (stderr,
	  "Possible security attack: double-check your crontab with 'crontab -l'\n");
	}
    if (tempstat1.st_dev != tempstat2.st_dev
      ||  tempstat1.st_ino != tempstat2.st_ino
      ||  tempstat1.st_nlink != tempstat2.st_nlink
      ||  tempstat1.st_uid != tempstat2.st_uid
      ||  tempstat1.st_gid != tempstat2.st_gid)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Temp file '%s' has stat mismatch, possible security attack.\n",
	  tempname);
	(void) fprintf (stderr,
	  "Double-check your crontab with 'crontab -l'\n");
	}
    (void) unlink (tempname);
    return status;
    }

static int addtab (user, filename)	/* Add a cron table */
    reg5 char *		user;		/* Name of owner of the table */
    reg4 char *		filename;	/* Name of file containing the table */
    {
    reg3 int		ch;		/* Character being copied */
    reg1 FILE *		infile;		/* Stdio pointer for reading file */
    struct stat		instat;		/* Stat buffer for input file */
    reg2 FILE *		outfile;	/* Pointer for writing cron copy */
    struct stat		outstat;	/* Stat buffer for output file */
#ifdef LONGNAMES
#define CRONTEMP	"ctXXXXXXXXXXXXX" /* Base for making temp file names */
#else /* LONGNAMES */
#define CRONTEMP	"ctXXXXXXXX"	/* Base for making temp file names */
#endif /* LONGNAMES */
    char		tempname[sizeof CRONTEMP]; /* Space for tmp file name */

    (void) strcpy (tempname, CRONTEMP);	/* Set up a temp file name */
    (void) tempfile (tempname);		/* Create the temp file name */
    if (filename[0] == '-'  &&  filename[1] == '\0')
	{
	infile = stdin;
	if (stat (user, &instat) < 0)	/* Try to duplicate existing modes */
	    instat.st_mode = CRONMODE;	/* Use default if all else fails */
	setlibdir ();
	(void) chdir (CRONTABDIR);
	}
    else
	{
	if (access (filename, 4) < 0)
	    {
	    (void) fprintf (stderr, "Can't access %s: %s\n", filename,
	      errmsg (0));
	    return 1;
	    }
	if ((infile = fopen (filename, "r")) == NULL)
	    {
	    (void) fprintf (stderr, "Can't access %s: %s\n", filename,
	      errmsg (0));
	    return 1;
	    }
	if (access (filename, 4) < 0)
	    {
	    (void) fprintf (stderr, "Can't access %s: %s\n", filename,
	      errmsg (0));
	    (void) fclose (infile);
	    return 1;
	    }
	setlibdir ();
	(void) chdir (CRONTABDIR);

	(void) fstat (fileno (infile), &instat);
	if (userpwent->pw_uid != instat.st_uid
	  &&  userpwent->pw_uid != ROOTUID)
	    {
	    /*
	     * Protect against timing attacks by refusing to read a
	     * crontab that's not owned by the user, unless the user
	     * is root.
	     */
	    (void) fprintf (stderr,
	      "Sorry, you must be the owner of the crontab you are submitting\n");
	    (void) fclose (infile);
	    return 1;
	    }
	if (instat.st_size == 0)	/* Treat null input as a delete req. */
	    {
	    (void) fclose (infile);
	    (void) chdir (LIBDIR);
	    return removetab (user);
	    }
	if (stat (user, &outstat) >= 0)	/* Check for identical files */
	    {
	    if (instat.st_dev == outstat.st_dev
	      &&  instat.st_ino == outstat.st_ino) /* Identical? */
		{
		(void) chdir (LIBDIR);	/* Yes, tell cron but don't copy */
		(void) fclose (infile);
		if (sendcommand ('a', user, 1) == SEND_ERROR)
		    return 1;
		else
		    return 0;
		}
	    }
	}

    if ((outfile = fopen (tempname, "w")) == NULL)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr,
	  "Can't create temp file '%s' to hold crontab: %s\n",
	  tempname, errmsg (0));
	(void) fclose (infile);
	return 1;
	}
    /*
     * Make the user file's ownership and modes correct.
     */
    if (chown (tempname, userpwent->pw_uid, userpwent->pw_gid) < 0)
	{
	(void) fprintf (stderr, "crontab:  couldn't change owner of %s: %s\n",
	  tempname, errmsg (0));
	(void) fclose (infile);
	(void) fclose (outfile);
	(void) unlink (tempname);
	(void) chdir (LIBDIR);
	return 1;
	}
    if (chmod (tempname, (instat.st_mode | MUSTHAVE) & ~CANTHAVE) < 0)
	{
	(void) fprintf (stderr, "crontab:  couldn't change mode of %s: %s\n",
	  tempname, errmsg (0));
	(void) fclose (infile);
	(void) fclose (outfile);
	(void) unlink (tempname);
	(void) chdir (LIBDIR);
	return 1;
	}
    /*
     * We've opened the prototype file and the copy in the spool directory.
     * Copy the prototype to the spool directory.  Note that we checked
     * earlier to make sure the files aren't identical.
     */
    outstat.st_size = 0;
    while ((ch = getc (infile)) != EOF)
	{
	(void) putc (ch, outfile);
	outstat.st_size++;
	}
    (void) fclose (infile);
    (void) fclose (outfile);
    /*
     * If the file came from standard input, it may have been of zero
     * length.  In that case we must remove the file we just created.
     */
    if (outstat.st_size == 0)
	{
	(void) unlink (tempname);
	(void) chdir (LIBDIR);
	return removetab (user);
	}
    /*
     * Replace the previous version of the user's crontab (if any) with the
     * current version.
     */
    if (unlink (user) < 0  &&  errno != ENOENT)
	{
	(void) fprintf (stderr,
	  "crontab:  couldn't unlink old cron table for %s: %s\n",
	  user, errmsg (0));
	(void) unlink (tempname);
	(void) chdir (LIBDIR);
	return 1;
	}
#ifdef HASRENAME
    if (rename (tempname, user) < 0)
	{
	(void) fprintf (stderr, "crontab:  couldn't rename %s to %s: %s\n",
	  tempname, user, errmsg (0));
	(void) unlink (tempname);
	(void) chdir (LIBDIR);
	return 1;
	}
#else
    if (link (tempname, user) < 0)
	{
	(void) fprintf (stderr, "crontab:  couldn't rename %s to %s: %s\n",
	  tempname, user, errmsg (0));
	(void) unlink (tempname);
	(void) chdir (LIBDIR);
	return 1;
	}
    else if (unlink (tempname) < 0)
	{
	(void) fprintf (stderr,
	  "crontab:  couldn't unlink %s after rename:  %s\n",
	  tempname, errmsg (0));
	(void) chdir (LIBDIR);
	return 1;
	}
#endif
    (void) chdir (LIBDIR);
    /*
     * Now tell the cron daemon to read the file.
     */
    if (sendcommand ('a', user, 1) == SEND_ERROR)
	return 1;
    else
	return 0;
    }

static int showlog (user)		/* Print cron log */
    reg5 char *		user;		/* User to show info about */
    {
    char		buf[40];	/* Buffer for fronts of lines */
    reg1 char *		bufp;		/* Pointer into buf */
    reg2 int		ch;		/* Character read from file */
    reg3 FILE *		logfile;	/* Stdio access to the cron log file */
    reg4 int		prline;		/* Print this line if NZ */

    if (strcmp (user, "root") == 0)
	{
	if (getuid () != 0)
	    {
	    (void) fprintf (stderr, "Security violation.\n");
	    return 1;
	    }
	user = NULL;
	}
    setlibdir ();
    logfile = fopen (CRONLOG, "r");
    if (logfile == NULL)
	{
	(void) fprintf (stderr, "Couldn't open %s/%s: %s\n", LIBDIR, CRONLOG,
	  errmsg (0));
	return 1;
	}
    while (fgets (buf, sizeof buf, logfile) != NULL)
	{
	prline = 0;
	if (user == NULL)
	    prline = 1;
	else if (buf[0] == '!')		/* Cron statistics line? */
	    {
	    /* Format is "!  user info" or "! *** all-user-info" */
	    if (buf[2] == '*')
		prline = 1;
	    else if ((bufp = strchr (buf + 3, ' ')) != NULL)
		{
		*bufp = '\0';
		if (strcmp (buf + 3, user) == 0)
		    prline = 1;
		*bufp = ' ';
		}
	    }
	else if (buf[0] == '<'  ||  buf[0] == '>') /* Proc start/end line? */
	    {
	    /* Format is "<  user info" */
	    bufp = strchr (buf + 3, ' ');
	    if (bufp)
		{
		*bufp = '\0';
		if (strcmp (buf + 3, user) == 0)
		    prline = 1;
		*bufp = ' ';
		}
	    }
	if (prline)
	    {
	    fputs (buf, stdout);
	    if (buf[strlen (buf) - 1] != '\n')
		{			/* If that wasn't all, do the rest */
		while ((ch = getc (logfile)) != EOF)
		    {
		    putchar (ch);
		    if (ch == '\n')
			break;
		    }
		}
	    }
	}
    (void) fclose (logfile);
    return 0;
    }
