#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/deny.c,v 2.5 1995/05/11 05:05:21 geoff Exp $";
#endif

/*
 * $Log: deny.c,v $
 * Revision 2.5  1995/05/11 05:05:21  geoff
 * Minor RCS header changes.
 *
 * Revision 2.4  87/03/31  22:32:29  geoff
 * Make the include of string.h configurable.
 * 
 * Revision 2.3  87/01/10  19:15:13  geoff
 * Fix a lint complaint.
 * 
 * Revision 2.2  86/11/29  19:51:04  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.1  86/04/13  17:34:43  geoff
 * Initial revision
 * 
 */

#include "config.h"
#include <stdio.h>
#include "pstring.h"

int		allowed ();	/* Decide whether a user is allowed access */
static int	isin ();	/* See if a user is listed in a file */

int allowed (user, priv)	/* Decide whether a user is allowed access */
    char *		user;	/* Name of user being tested */
    char *		priv;	/* Name of the privilege he's trying to use */
    {
    int			allowstat; /* Result of looking in allow file */

    allowstat = isin (user, priv, ALLOW);

    if (allowstat > 0)
	return 1;

    if (isin (user, priv, DENY) > 0)
	return 0;

    /*
     * The user is not in either the allow or the deny files.  Make a
     * default decision based on the following rules:
     *
     * (1) If the allow file exists, the system manager is assumed to be
     * interested in limiting access to a select group, so the default
     * is to deny access (otherwise there is no point in having an allow
     * file).
     *
     * (2) If no allow file exists, the system manager is assumed to be
     * interested in allowing access to most people, with a few denied
     * access via the deny file.  So we allow access.
     */
    if (allowstat == 0)
	return 0;		/* Allow exists but the user isn't in it */
    else
	return 1;		/* No allow file exists */
    }

/*
 * See if a user is listed in a file
 *
 * Returns 1 if in file, 0 if not, -1 if file can't be opened.
 */
static int isin (user, priv, suffix)
    reg1 char *		    user;	/* User to look for */
    char *		    priv;	/* Privilege file being looked in */
    char *		    suffix;	/* Suffix to add to file name */
    {
    char		    buf[512];	/* General scratch buffer */
    reg2 FILE *		    pfile;	/* Access to privilege file */

    (void) strcpy (buf, priv);		/* Build the file name */
    (void) strcat (buf, suffix);

    pfile = fopen (buf, "r");		/* Open the privilege file */
    if (pfile == NULL)
	return -1;			/* No such file */

    while (fgets (buf, sizeof buf, pfile) != NULL) /* Search for user */
	{
	buf[strlen (buf) - 1] = '\0';	/* Trim off trailing newline */
	if (strcmp (buf, user) == 0)	/* Does it match the user name? */
	    {
	    (void) fclose (pfile);
	    return 1;			/* Yes, return "is in" */
	    }
	}
    (void) fclose (pfile);

    return 0;				/* Not found */
    }
