#!/usr/bin/perl -w

#
# location_monitor.pl
# $Id: location_monitor.pl,v 1.1 2000/04/13 04:31:53 johnh Exp $
#
# Copyright (C) 1999 by John Heidemann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#



sub usage {
    print STDERR <<END;
usage: $0 [-D]

monitor and filter location, sending signals to lcron if it changes

With no arguments, runs once.
-D puts it in "daemon" mode where it waits and runs every SIGIO.
END
    exit 1;
}

use strict;

use Getopt::Long;
&usage if ($#ARGV >= 0 && $ARGV[0] eq '-?');
my(%opts);
&GetOptions(\%opts, qw(D R));

my($BINDIR) = '/usr/local/lib/xcron/bin';
my($LIBDIR) = '/usr/local/lib/xcron/lib';


if (defined($opts{'D'})) {
    my($f) = fork;
    die "$0: fork failed\n" if (!defined($f));
    exit 0 if ($f != 0);
};

&netreport if (defined($opts{'R'}));
&daemon_mode if (defined($opts{'D'}));
&check();
exit 0;


sub netreport {
    open (PIDFILE, ">/var/run/netreport/$$") || die "$0: cannot create /var/run/netreport/$$.\n";
    print PIDFILE "$0\n", scalar localtime, "\n";
    close PIDFILE;
}


sub check {
    printf "check\n";

    my($netstat);
    foreach (qw(/usr/bin/netstat /bin/netstat /etc/netstat /usr/ucb/netstat /dev/null)) {
	$netstat = $_, last if (-x $_);
    };
    die "$0: cannot find netstat\n" if ($netstat eq '/dev/null');

    my(@gws) = ();
    my($gw);
    open(NS, "$netstat -rn |") || die "$0: $netstat failed.\n";
    open(GW, ">$LIBDIR/events/gateways") || die "$0: cannot open '$LIBDIR/events/gateways'.\n";
    chmod(0644, "$LIBDIR/events/gateways");
    while (<NS>) {  
	my(@f) = split;
	my($addr);
	foreach $addr ($f[0], $f[1]) {
	    if ($addr !~ /^(0\.0\.0\.0|Gateway|IP)/) {
		print GW "$addr\n";
		push(@gws, $addr);
	    };
	};
    };
    close NS;
    close GW;

    my(%gw_to_place) = ();
    if (-f "$LIBDIR/netstat_mapping") {
	open(MAP, "<$LIBDIR/netstat_mapping") || warn "$0: cannot open '$LIBDIR/netstat_mapping'.\n";
	while (<MAP>) {
	    my(@f) = split;
	    my($gw) = shift @f;
	    $gw_to_place{$gw} = [] if (!defined($gw_to_place{$gw}));
	    push (@{$gw_to_place{$gw}}, @f);
	}
	close MAP;
    };
    open(PL, ">$LIBDIR/events/places") || die "$0: cannot open '$LIBDIR/events/places'.\n";
    chmod(0644, "$LIBDIR/events/places");
    foreach $gw (@gws) {
	if (defined($gw_to_place{$gw})) {
	    my($pl);
	    foreach $pl (@{$gw_to_place{$gw}}) {
	    	print PL "$pl\n";
	    };
	};
    };
    close PL;

    system("$BINDIR/tellxcron", "-L");
}

sub daemon_exit_error {
    exit 1;
}

sub daemon_exit_noerror {
    exit 0;
}

sub daemon_mode {
    $SIG{'IO'} = \&check;
    $SIG{'HUP'} = \&daemon_exit_noerror;
    $SIG{'INT'} = \&daemon_exit_error;
    $SIG{'TERM'} = \&daemon_exit_error;

    for (;;) {
	sleep 6000;
    };
    die "$0: error, daemon mode ecaped infinte loop.\n";
}

