#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/mainloop.c,v 2.16 2000/04/13 03:19:35 johnh Exp $";
#endif

/*
 * $Log: mainloop.c,v $
 * Revision 2.16  2000/04/13 03:19:35  johnh
 * fixes for several bugs: loc-specific at jobs running multiple times and at-job reading, and geoffs mis-applied Y2K patch
 *
 * Revision 2.15  2000/04/10 15:55:18  johnh
 * close bug in checkloc fixed
 *
 * Revision 2.14  2000/02/08 08:56:42  geoff
 *
 * Integrate various changes from Geoff's branch into the now-master CVS
 * repository.  As part of this integration, I cleaned up some, but not
 * all, of the inconsistent coding style introduced by the LOC_CRON
 * changes.  Important and substantive changes include:
 *
 *     Configure	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     Makefile.SH	Tiny cleanup to an expr command.
 *     config.SH	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     errmsg.c	Add errlist_in_stdio support.
 *     mainloop.c	Use setjmp/longjmp to ensure that we are blown off the
 * 		select when we catch an alarm, even if we are running
 * 		on a system that has reliable signals.  Also add
 * 		reliable-signal support that properly resets the
 * 		signal mask in the longjmp case.
 *     parse.c	Fix a bug that could cause crontabs to be accepted
 * 		with the wrong UID.  Be a bit more careful about
 * 		parsing years.
 *     runcmd.c	Add setjmp/longjmp and reliable-signal support.  (See
 * 		mainloop.c).
 *     xcron.c	Use setitimer to make absolutely sure that all alarms
 * 		are disabled at startup.
 *
 * Revision 2.13  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.12  1998/07/30 20:42:26  geoff
 * Many changes, accumulated over far too much time: add BSD socket
 * support; don't do catchups from now to now.
 *
 * Revision 2.11  1995/11/27  01:59:41  geoff
 * Add ONLYCATCHUP, MAXSLEEP, and MINUPDLAST support.
 *
 * Revision 2.10  1995/05/11  05:05:32  geoff
 * Support variable signal types.  Use a different variable to allocate
 * file-name space.  Only garbage collect after the file-deletion loop
 * completes.  Clean out the fifo before exiting.
 *
 * Revision 2.9  87/03/31  22:33:09  geoff
 * Make fifofd external.  Add code to report when the two catchup phases
 * complete.
 * 
 * Revision 2.8  86/12/01  14:45:42  geoff
 * Fix some lint complaints.
 * 
 * Revision 2.7  86/11/29  19:52:03  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.6  86/11/16  00:58:20  geoff
 * Unlink the lock file in cleanup;  this makes orderly exits look nicer.
 * 
 * Revision 2.5  86/04/17  21:55:56  geoff
 * Add the user name to the 'deleted' log message so that the -L switch
 * to crontab can find it.
 * 
 * Revision 2.4  86/04/16  00:27:10  geoff
 * Fix a core-dump bug in the last change to the cron-exit code
 * 
 * Revision 2.3  86/04/13  13:32:15  geoff
 * Add code to protect against overflow of the argument to alarm(2).  Also
 * add code to handle deletion of at jobs (just a message change).
 * 
 * Revision 2.2  86/04/02  00:01:54  geoff
 * Move the checkfifo call into waituntil, and add an updlast call so the
 * last-run time is always up to date.
 * 
 * Revision 2.1  86/04/01  21:51:48  geoff
 * Initial revision.
 * 
 */

#include "cron.h"
#include <errno.h>
#include <setjmp.h>
#include <stdio.h>
#include <sys/dir.h>		/* To get MAXNAMLEN */
#ifdef BSDSOCKET
#include <sys/types.h>
#include <sys/time.h>
#endif /* BSDSOCKET */

#ifdef LOC_CRON
#ifndef sun
#include <regex.h>   /* for re_{comp,exec} */
#endif
#include <string.h>    /* for strtok */
#include <dirent.h>
#endif /* LOC_CRON */

time_t		catchup ();	/* Run stuff that has to catch up to now */
static void	re_sort ();		/* Re-sort entries after catchup */
void		cron ();	/* Main loop of the cron daemon */
static time_t	waituntil ();	/* Wait for work interval to arrive */
static void	updlast ();	/* Update the last-ran record */
SIG_TYPE	fifocatch ();	/* Catch READFIFO signals */
SIG_TYPE	deathcatch ();	/* Catch deadly signals */
static void	checkfifo ();	/* Check FIFO for commands */
#ifdef LOC_CRON
void     checkloc ();    /* Check for location */
#endif
static int	fifogetc ();	/* Get a character from the FIFO */
static void	cleanup ();	/* Clean up temp files before exit */

struct utimbuf
    {
    time_t	actime;		/* Isn't it nice of AT&T to give us */
    time_t	modtime;	/* ..header files for everything? */
    };

static short	fifohasdata = 1; /* NZ if data is in the FIFO */
				/* ..we assume yes at startup mostly */
				/* ..because it's a way to get the */
				/* ..fifo opened */
#ifdef LOC_CRON
static short    haslocchanged = 1;
				/* NZ if location has changed (?) */
#endif
int		fifofd = -1;	/* File descriptor of the fifo */

extern struct cronfile *
		addfile ();	/* Add a file to the cronfile list */
extern unsigned	alarm ();	/* Set alarm clock timeout */
extern int	alarm_caught;	/* NZ if an alarm has been caught */
extern jmp_buf	alarm_jb;	/* Jump buffer for catching alarms */
extern int	alarm_jb_valid;	/* NZ if alarm_jb has valid value */
extern void	collect ();	/* Garbage collection routine */
#ifdef DEBUG
extern short	debug;		/* Nonzero to get debugging output */
#endif
extern void	delfile ();	/* Delete entries from a given file */
#ifdef DUMP
extern void	dumpall ();	/* Dump all data structures */
#endif
extern struct cronent
		enthead;	/* Head of cronent chain */
extern char *	errmsg ();	/* Return an error message string */
extern int	errno;		/* System error number */
extern void	_exit ();	/* Exit without buffer flushing */
extern void	exit ();	/* Exit program */
extern struct cronfile
		filehead;	/* Head of cronfile chain */
extern time_t	findandrun ();	/* Find runnable cmds and run them */
extern void	insent ();	/* Insert entry into the entry list */
extern time_t	nextrun ();	/* Figure out when entry runs next */
#ifdef DEBUG
extern char *	printable ();	/* Return printable version of char */
#endif
extern struct childproc
		prochead;	/* Head of child process list */
extern short	re_collect;	/* NZ if collection is needed later */
extern void	readall ();	/* Read in all files */
extern void	readfile ();	/* Read a file into cronent lists */
extern void	readlimits ();	/* Read the user limits file */
extern void	runqueues ();	/* Run the commands on the queues */
extern unsigned	sleep ();	/* Suspend process for a while */
extern void	sync ();	/* System call to sync the disks */
extern time_t	syncinterval;	/* Interval between syncs, if any */
extern time_t	time ();	/* Return system time */
extern char *	timenonl ();	/* Return a time w/o newline */
extern int	waitfor ();	/* Wait for a process to finish */

/*
 * Catch up to the system's idea of when it is.  Return value is cron's
 * new idea of when it is, which (at least briefly) will match the
 * system time.
 */
time_t catchup (fromwhen)		/* Do any catching-up required */
    reg4 time_t		fromwhen;	/* Time we last ran successfully */
    {
    reg1 struct cronent * entp;		/* Current crontab entry */
    int			kidpid;		/* Process ID of spawned child */
    reg2 time_t		nextdue;	/* When the next command is due */
    int			saveerr;	/* Place to save errno b4 clobber */
    reg3 time_t		sysnow;		/* System's idea of the current time */

    sysnow = time ((long *) NULL);
#ifdef LOC_CRON
    /*
     * For some reaon lcron removes this optimization
     * (maybe to make sure we run jobs when we change location?).
     */
#else /* ! LOC_CRON */
    if (sysnow == fromwhen)
	return;
#endif /* LOC_CRON */
    /*
     * Clear the HASRUN bits.
     */
    for (entp = enthead.ce_next;  entp != &enthead;  entp = entp->ce_next)
	entp->ce_flags &= ~EF_HASRUN;
    /*
     * The fprintf must be done in two parts, because timenonl returns
     * a pointer to a static area.
     */
    (void) fseek (stderr, 0L, 2);
    (void) fprintf (stderr, "! *** cron catchup from %s",
      timenonl (fromwhen));
    (void) fprintf (stderr, " to %s\n", timenonl (sysnow));
    /*
     * Catchup is performed in two phases.  In the first phase, we run all
     * SPAWN and WAIT commands.  In the second phase, we fork into background
     * and run all SEQSPAWN commands.
     *
     * The commands due at 'fromwhen' have already been run, so we
     * start at the next minute boundary.
     */
    fromwhen += MINUTE - (fromwhen % MINUTE);
    /*
     * If you are wondering why we don't just call 'findandrun' with an
     * argument of 'sysnow', it's because that would only run the RUNALL
     * stuff once.  Otherwise, we could do that.
     */
    for (nextdue = fromwhen;  nextdue <= (sysnow = time ((long *) NULL));  )
	nextdue =
	  findandrun (nextdue, EF_SPAWN | EF_WAIT, EF_SEQSPAWN | EF_ATJOB);
    nextdue = time ((long *) NULL);
    (void) fseek (stderr, 0L, 2);
    (void) fprintf (stderr, "! *** spawn/wait catchup complete at %s\n",
      timenonl (nextdue));
    /*
     * We have now run all SPAWN and WAIT commands.  Run SEQSPAWN's in
     * background.  Note that we don't update 'sysnow', even if we go into
     * the sleep loop.  This causes the SEQSPAWN command list to be the
     * same as the list used for SPAWN/WAIT.
     */
    while ((kidpid = fork ()) == -1  &&  errno == EAGAIN)
	(void) sleep (5);
    if (kidpid == -1)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "catchup:  couldn't fork:  %s\n",
	  errmsg (saveerr));
	}
    else if (kidpid == 0)		/* Child runs the commands */
	{
#ifdef DEBUG
	if (debug < 15)			/* Low debug levels don't go in kid */
	    debug = 0;
#endif
	for (nextdue = fromwhen;  nextdue <= sysnow;  )
	    nextdue = findandrun (nextdue, EF_SEQSPAWN, EF_ATJOB);
	nextdue = time ((long *) NULL);
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "! *** seqspawn catchup complete at %s\n",
	  timenonl (nextdue));
	_exit (0);
	/* NOTREACHED */
	}
    /*
     * The parent has to re-arrange the SEQSPAWN jobs in the queue before
     * it is done.
     */
    re_sort (sysnow);
    return sysnow;
    }

static void re_sort (sysnow)		/* Re-sort entries after catchup */
    reg3 time_t			sysnow;	/* Current system time */
    {
    reg1 struct cronent *	entp;	/* Current crontab entry */
    reg2 struct cronent *	nentp;	/* Following entry */

    /*
     * After catchup, the entry queue contains SEQSPAWN jobs that
     * are in the wrong position, plus at jobs that are in the
     * right position but that have times earlier than 'sysnow'.
     * The following loop corrects both of these problems.
     */
    for (entp = enthead.ce_next;
      entp != &enthead  &&  entp->ce_rtime <= sysnow;
      entp = nentp)
	{
	nentp = entp->ce_next;
	if (entp->ce_flags & (EF_SEQSPAWN | EF_DELETED))
	    {
	    entp->ce_prev->ce_next = entp->ce_next; /* Delink it */
	    entp->ce_next->ce_prev = entp->ce_prev;
	    entp->ce_rtime = nextrun (sysnow, entp);
	    insent (entp);
	    }
	else if (entp->ce_flags & EF_ATJOB)
	    entp->ce_rtime = sysnow;
#ifdef LOC_CRON
	else if (entp->ce_flags & EF_ONLY_ON_MOVE) {
	    /*
	     * The only-on-move jobs also seem to end up in this queue,
	     * probably only when their locations don't allow
	     * them to run.
	     * I'm not sure how I should fix this problem since
	     * I don't understand everything Geoff's doing here.
	     * I'll pretned I do and treat them like atjobs
	     * until he tells me better.
	     */
	     entp->ce_rtime = sysnow;
	}
#endif /* LOC_CRON */
#ifdef DEBUG
	else
	    {
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "catchup: Unexpected out-of-position entry:\n");
	    dumpent (entp);
	    entp->ce_rtime = sysnow;
	    }
#endif
	}
    }

void cron (ournow)			/* Main loop of the cron daemon */
    reg2 time_t			ournow;	/* Our idea of the current time */
    {
    reg1 time_t			nextdue; /* When the next command is due */

    for (  ;  ;  )
	{
	int ignore = EF_ONLYCATCHUP;
#ifdef LOC_CRON
	ignore |= EF_ONLY_ON_MOVE_FIRED;
#endif /* LOC_CRON */
	nextdue = findandrun (ournow, 0, ignore);
					/* Queue jobs that are due */
	runqueues ();			/* Actually run jobs */
	updlast (ournow, (time_t) 0);	/* Record when we last did something */
	if (syncinterval)		/* Sync even if it's not time */
	    sync ();
	if (re_collect)
	    collect ();			/* Do a garbage collection */
	ournow = waituntil (ournow, nextdue);
	}
    }

	/* Wait for a work interval to arrive */
static time_t waituntil (ournow, duetime)
    reg5 time_t			ournow;		/* Our idea of the time */
    reg2 time_t			duetime;	/* When the work is due */
    {
#ifdef BSDSOCKET
    int				nfds;		/* Number of ready fds */
#ifdef RELIABLE_SIGNALS
    int				oldsigs;	/* Previous signal mask */
#endif /* RELIABLE_SIGNALS */
    fd_set			readfds;	/* FDs ready for read */
#endif /* BSDSOCKET */
    reg1 time_t			sysnow;		/* System's idea of the time */
    reg3 unsigned int		sleeptime;	/* Time to set alarm for */
#ifdef BSDSOCKET
#ifdef HASSIGPROCMASK
    sigset_t			unblockset;	/* Signals to be unblocked */
#endif /* HASSIGPROCMASK */
#endif /* BSDSOCKET */
    reg4 int			waitval;	/* Return value from waitfor */
    
    DBPRINT1 (15, "waituntil from %s", timenonl (ournow));
    DBPRINT1 (15, " to %s\n", timenonl (duetime));
    for (  ;  ;  )
	{
	/*
	 * If a crontab is submitted, changed, or deleted, the
	 * "fifohasdata" flag will be set, and we call "checkfifo"
	 * to do what the fifo wants.  After checking the fifo, we
	 * pretend "duetime" has arrived and return;  this allows
	 * the main loop to update duetime if earlier work has
	 * arrived.
	 */
	if (fifohasdata)
	    {
	    checkfifo (ournow);			/* Handle file list changes */
#ifdef LOC_CRON
	    /* User may have changed locations so refigure things */
	    checkloc();
#endif
	    DBPRINT0 (15, "waituntil: fifo data arrived\n");
	    return ournow;
	    }
#ifdef LOC_CRON
	if (haslocchanged)
	  {
	    DBPRINT0(15, "waituntil: calling checkloc \n");
	    checkloc();
	    return ournow;
	  }
#endif

	sysnow = time ((long *) NULL);
	/*
	 * NEEDSWORK: should run catch-up on spring daylight switch
	 */
	if (sysnow > ournow + CATCHUP		/* Time change? */
	  ||  sysnow < ournow - CATCHUP)
	    {
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr, "! *** time change %swards from %s",
	      (sysnow < ournow) ? "back" : "for", timenonl (ournow));
	    (void) fprintf (stderr, " to %s\n", timenonl (sysnow));
	    /*
	     * For backward changes, we have to reread all the files
	     * because some entries may have been deleted as being too
	     * old.  For forward changes of over a year, the year base
	     * may be wrong and again the files must be reread.  Time
	     * changes shouldn't happen very often, so the cost is
	     * acceptable.
	     */
	    if (sysnow < ournow  ||  sysnow > ournow + 365 * 24 * HOUR)
		readall (sysnow);		/* Rebuild entry list */
	    /*
	     * Catch up to the new current time.
	     */
	    ournow = catchup (sysnow < ournow ? sysnow : ournow);
	    DBPRINT0 (15, "waituntil: catchup has completed\n");
	    return ournow;			/* Let caller recalc duetime */
	    }
	else if (sysnow >= duetime)
	    {
	    DBPRINT1 (15, "waituntil: due time of %s has arrived",
	      timenonl (duetime));
	    DBPRINT1 (15, ", it is now %s\n", timenonl (sysnow));
	    return sysnow;
	    }
	else
	    {
	    /*
	     * There have been no time changes.  Wait for duetime to arrive.
	     */
	    updlast (sysnow, MINUPDLAST);
	    sleeptime = duetime - sysnow;
	    /*
	     * Never sleep too long, so that we can pick up time changes
	     * relatively soon.
	     */
	    if (sleeptime > MAXSLEEP)
		sleeptime = MAXSLEEP;
	    DBPRINT2 (6, "sleep for %d seconds, until %s\n",
	      sleeptime, timenonl (sysnow + sleeptime));
	    ournow = sysnow;
	    (void) alarm (sleeptime);
	    waitval = waitfor (0);	/* Wait for a process to die */
	    DBPRINT1 (15, "waituntil: waitfor returned %d\n", waitval);
	    if (waitval == 0)		/* Did a kid die? */
		{
		ournow += sleeptime - alarm (0);
		runqueues ();		/* See if a proc slot is now free */
		}
	    else
		{
#ifdef BSDSOCKET
		/*
		 * We use an alarm to blast us off of the select,
		 * rather than the built-in timeout capability,
		 * because that allows us to discover how long we were
		 * asleep.
		 */
		if (waitval < 0)
		    {
		    FD_ZERO (&readfds);
		    FD_SET (fifofd, &readfds);
#ifdef RELIABLE_SIGNALS
		    oldsigs = siggetmask ();
#endif /* RELIABLE_SIGNALS */
		    if (setjmp (alarm_jb))
			{
			nfds = 0;	/* An alarm has been caught */
			alarm_jb_valid = 0;
#ifdef HASSIGPROCMASK
			/*
			 * Since longjmp didn't return properly from
			 * the signal handler, a POSIX system will
			 * wind up with the alarm blocked.  So we have
			 * to unblock the signal again.
			 */
			sigemptyset (&unblockset);
			sigaddset (&unblockset, SIGALRM);
			sigprocmask (SIG_UNBLOCK, &unblockset,
			  (sigset_t *) NULL);
#endif /* HASSIGPROCMASK */
#ifdef RELIABLE_SIGNALS
			/*
			 * Since longjmp didn't return properly from
			 * the signal handler, a BSD system will wind
			 * up with the alarm blocked.  So we have to
			 * restore the old mask, which we have
			 * carefully saved, to unblock it.
			 */
			sigsetmask (oldsigs);
#endif /* RELIABLE_SIGNALS */
			}
		    else
			{
			alarm_jb_valid = 1;
			if (alarm_caught)
			    nfds = 0;
			else
			    nfds = select (fifofd + 1, &readfds,
			      (fd_set *) NULL, (fd_set *) NULL,
			      (struct timeval *) NULL);
					/* No kids:  wait for work */
			alarm_jb_valid = 0;
			}
		    }
		if (nfds > 0)
		    fifohasdata = 1;
#else /* BSDSOCKET */
		if (waitval < 0)
		    (void) pause ();	/* No kids:  pause for alarm instead */
#endif /* BSDSOCKET */
		ournow += sleeptime - alarm (0);
		}
	    }
	}
    }

static void updlast (when, interval)	/* Update the last-ran record */
    reg1 time_t			when;	/* When we ran last */
    reg2 time_t			interval; /* NZ to update only if last */
					/* ..update more than this time ago */
    {
    static time_t		errtime; /* Time of last error report */
    static struct utimbuf	lastran; /* Utime update struct */
					/* ..we depend on init to 0 */
    int				saveerr; /* Place to save errno b4 clobber */

    DBPRINT2 (14, "updlast (%s, %ld)", timenonl (when), interval);
    DBPRINT1 (14, ", lastran is %s\n", timenonl (lastran.modtime));
    when -= when % 60;
    if (interval != 0  &&  when < lastran.modtime + interval)
	{
	DBPRINT0 (15, "updlast: returning due to unsatisfied interval\n");
	return;
	}
    if (when >= lastran.modtime + MINUTE) /* Update once a minute at most */
	{
	lastran.modtime = lastran.actime = when;
	DBPRINT1 (15, "updlast: updating time to %s\n", timenonl (when));
	if (utime (TIMELOG, &lastran) < 0) /* Change file times */
	    {
	    if (when >= errtime + ERRORTIME) /* Only report every 10 minutes */
		{
		saveerr = errno;
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr,
		  "cron:  couldn't update times on %s:  %s\n",
		  TIMELOG, errmsg (saveerr));
		errtime = when;
		}
	    }
	}
    else
	DBPRINT0 (15,
	  "updlast: returning due to interval less than 1 minute\n");
    }

/* ARGSUSED */
SIG_TYPE fifocatch (sig)		/* Catch READFIFO signals */
    int			sig;		/* Signal that brought us here */
    {
    (void) signal (READFIFO, fifocatch);
    fifohasdata = 1;			/* Request FIFO read from main loop */
    }

SIG_TYPE deathcatch (sig)		/* Catch deadly signals */
    int			sig;		/* Signal that brought us here */
    {
    cleanup ();				/* Get rid of left-over temp files */
    (void) fseek (stderr, 0L, 2);
    (void) fprintf (stderr, "! *** cron terminated due to signal %d %s\n",
      sig, timenonl (time ((time_t *) NULL)));
    (void) signal (sig, SIG_DFL);	/* Make sure we have default setting */
    (void) kill (getpid (), sig);	/* Re-deliver signal to kill ourself */
    }

static void checkfifo (reftime)		/* Check FIFO for commands */
    time_t		reftime;	/* Time to reference new entries to */
    {
    reg3 int		ch;		/* Character read from FIFO */
    reg4 int		cmd;		/* Command to be issued */
    reg2 char *		cp;		/* Pointer into filename buffer */
    char		filename[MAXNAMLEN + 1]; /* Space for file name */
    reg1 struct cronfile * filep;	/* File table entry for file */

    DBPRINT0 (10, "checkfifo ()\n");
    fifohasdata = 0;
    while ((cmd = fifogetc ()) != EOF)	/* Loop while there's work */
	{
	/* We have a command.  Skip the blank after it. */
	ch = fifogetc ();
	if (ch == EOF)
	    {
	    (void) fprintf (stderr, "checkfifo:  unexpected EOF on FIFO\n");
	    return;
	    }
	else if (ch != ' ')
	    {
	    (void) fprintf (stderr,
	      "checkfifo:  sequence error, skipping bytes\n");
	    while ((ch = fifogetc ()) != '\n')
		if (ch == EOF)
		    return;
	    continue;
	    }
	/* Now read the filename itself, up to a newline, and trimming */
	for (cp = filename;  (ch = fifogetc ()) != '\n';  )
	    {
	    if (ch == EOF)
		{
		(void) fprintf (stderr,
		  "checkfifo:  sequence error (missing bytes)\n");
		return;
		}
	    *cp = ch;			/* Save next filename byte */
	    if (cp != &filename[MAXNAMLEN])
		cp++;
	    }
	*cp = '\0';			/* Terminate filename */
	switch (cmd)
	    {
	    case 'A':
		filep = addfile (filename, FF_ATJOB);
		if (filep != NULL)	/* Force reread and garbage-coll */
		    readfile (filep, 1, 1, reftime);
		break;
	    case 'a':			/* Add or reread a file */
		filep = addfile (filename, 0);
		if (filep != NULL)	/* Force reread and garbage-coll */
		    readfile (filep, 1, 1, reftime);
		break;
#ifdef DUMP
	    case 'D':
		dumpall ();		/* Do a dump */
		break;
#endif
	    case 'd':			/* Delete a file */
		for (filep = filehead.cf_next;
		  filep != &filehead;
		  filep = filep->cf_next)
		    {
		    if (strcmp (filename, filep->cf_name) == 0)
			{		/* Found what to delete */
			filep->cf_flags |= (FF_MARKDEL | FF_DELETED);
			(void) fseek (stderr, 0L, 2);
			(void) fprintf (stderr,
			  "!  %s %s/%s deleted %s\n",
			  filep->cf_user->cu_name + EL_LOGNAME,
			  (filep->cf_flags & FF_ATJOB)
			    ? ATJOBDIR : CRONTABDIR,
			  filep->cf_name,
			  timenonl (time ((long *) NULL)));
			delfile (filep); /* Get rid of it */
			}
		    }
		collect ();		/* Garbage-collect deleted entries */
		break;
	    case 'e':			/* Exit cron */
		while ((ch = fifogetc ()) != EOF)
		    ;			/* Clean out the fifo */
		cleanup ();		/* Get rid of outstanding temp files */
		updlast (reftime, (time_t) 0);
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr,
		  "! *** cron terminated by request %s\n", timenonl (reftime));
		exit (0);
		/* NOTREACHED */
#ifdef LOC_CRON
	    case 'L':
		haslocchanged = 1;
                break;
#endif /* LOC_CRON */
#ifdef DEBUG
	    case 'x':			/* Change debugging level */
		debug = atoi (filename);
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr, "Debugging level changed to %d\n",
		  debug);
		break;
#endif
	    case 'u':			/* Reread user limits file */
		readlimits ();
		break;
	    }
	}
    }

static int fifogetc ()			/* Get a character from the FIFO */
    {
    static char		fifobuf[512];	/* Buffer for FIFO data */
    static char *	fifop;		/* Pointer to next char in fifobuf */
    static int		fifosize = 0;	/* Amount of data in FIFO buffer */
    static time_t	msgtime = 0;	/* Time last msg was given */
    int			saveerr;	/* Place to save errno b4 clobber */

    if (fifofd < 0)			/* Is fifo open yet? */
	{				/* ..if lt no, must open it */
	DBPRINT0 (1, "fifogetc:  opening fifo\n");
	fifofd = openfifo (FIFONAME, 0); /* Open the fifo */
	if (fifofd < 0  &&  errno == ENOENT) /* Failure due to no such FIFO? */
	    {
	    if (mkfifo (FIFONAME, 0600) < 0) /* Try to make one */
		{
		if (time ((long *) NULL) >= msgtime + ERRORTIME)
		    {
		    saveerr = errno;
		    (void) fseek (stderr, 0L, 2);
		    (void) fprintf (stderr,
		      "fifogetc:  couldn't make %s: %s\n",
		      FIFONAME, errmsg (saveerr));
		    msgtime = time ((long *) NULL);
		    }
		return EOF;
		}
	    fifofd = openfifo (FIFONAME, 0); /* Try again to open the file */
	    }
	}
    if (fifofd < 0)			/* Still not open? */
	{
	if (time ((long *) NULL) >= msgtime + ERRORTIME)
	    {
	    saveerr = errno;
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr, "fifogetc:  couldn't open %s: %s\n",
	      FIFONAME, errmsg (saveerr));
	    msgtime = time ((long *) NULL);
	    }
	return EOF;
	}
    if (fifosize <= 0)			/* If buffer is empty, fill it */
	{
	fifosize = readfifo (fifofd, fifobuf, sizeof fifobuf);
	fifop = fifobuf;
	}
    if (fifosize <= 0)			/* If buffer still empty, return EOF */
	{
	DBPRINT0 (15, "fifogetc:  returning EOF\n");
	return EOF;
	}
    else
	{
	--fifosize;			/* Count char we're gonna deliver */
	DBPRINT2 (15, "fifogetc:  returning 0x%x ('%s')\n",
	  *fifop & CHARMASK, printable (*fifop));
	return *fifop++ & CHARMASK;	/* and deliver it */
	}
    }

static void cleanup ()			/* Clean up temp files before exit */
    {
    reg1 struct childproc *	procp;	/* Current child process entry */

    for (procp = prochead.cp_next;
      procp != &prochead;
      procp = procp->cp_next)
	{
	procp->cp_fbase[TEMPTYPELOC] = 'o';
	(void) unlink (procp->cp_fbase);
	procp->cp_fbase[TEMPTYPELOC] = 'e';
	(void) unlink (procp->cp_fbase);
	if (procp->cp_ent->ce_input)
	    {
	    procp->cp_fbase[TEMPTYPELOC] = 'i';
	    (void) unlink (procp->cp_fbase);
	    }
	}
    (void) unlink (LOCKFILE);
    }

#ifdef LOC_CRON
/*
 *
 */
int
loc_are_we_there_yet(entp)
	struct cronent *entp;
{
	if (entp->ce_location == NULL)
		return 0;
	if (entp->ce_flags & EF_ONLY_ON_MOVE) {
		return ((entp->ce_flags & EF_IS_AT_LOCATION) && !(entp->ce_flags & EF_WAS_AT_LOCATION)) ? EF_ONLY_ON_MOVE|EF_IS_AT_LOCATION : 0;
	};
	return (entp->ce_flags & EF_IS_AT_LOCATION) ? EF_IS_AT_LOCATION : 0;
}

struct location {
	struct location *next;
	char *name;
};

struct location *
checkloc_file(name, tail)
	char *name;
	struct location *tail;
{
#define LINELEN 256
	char line[LINELEN];
	char *last;
	FILE *f;

	if (NULL == (f = fopen(name, "r")))
		return tail;
	while (fgets(line, LINELEN, f)) {
		line[LINELEN-1] = 0;
		last = &line[strlen(line)-1];
		if (*last == '\n')
			*last = 0;
		if (NULL == (tail->next = malloc(sizeof(struct location))))
			break;
		tail = tail->next;
		tail->name = strdup(line);
		tail->next = NULL;
	};
	if (0 != fclose(f)) {
		fprintf(stderr, "checkloc_file: cannot close file.\n");
	};
	return tail;
}

/*
 * Figure out the location by parsing LIB/events
 */
void 
checkloc()
{
	struct cronent * ent;
	char *msg;
	int re_matches;
	DIR *dir = NULL;
	struct dirent *de;
	struct location location_head;
	struct location *location_tail = &location_head;
	struct location *location, *next_location;
	struct stat sb;

	location_head.next = NULL;
	location_head.name = NULL;

	haslocchanged = 0;

	/*
	 * Read and parse the locations
	 */
	if (-1 == chdir(LIBDIR)) {
		DBPRINT0(2, "checkloc: cannot chdir LIBDIR\n");
		return;
	};
	if (-1 == chdir("events")) {
		DBPRINT0(2, "checkloc: cannot chdir events\n");
		goto done;
	};
	if (NULL == (dir = opendir("."))) {
		DBPRINT0(2, "checkloc: cannot readdir .\n");
		goto done;
	};
	while (de = readdir(dir)) {
		char *last = &de->d_name[strlen(de->d_name)-1];
		if (*last == '~' || *last == '-') {
			/* ignore emacs backup files and minus files */
			continue;
		};
		if (-1 == stat(de->d_name, &sb)) {
			DBPRINT1(2, "checkloc: cannot stat %s\n", de->d_name);
			continue;
		};
		if (sb.st_uid != ROOTUID || sb.st_gid != ROOTGID || (sb.st_mode & 022) != 0) {
			DBPRINT1(2, "checkloc: security loose on %s (ownership or perms)\n", de->d_name);
			continue;
		};
		if (S_ISDIR(sb.st_mode))
			continue;
		location_tail = checkloc_file(de->d_name, location_tail);
	};
	closedir(dir);

	/*
	 * Now scan through all the entries.
	 */
	for (ent=enthead.ce_next;ent != &enthead;ent=ent->ce_next) {
		if (ent->ce_location == NULL)
			continue;
		if((msg = re_comp(ent->ce_location)) != NULL) {
			printf("Error '%s' in compiling regular expression %s\n", msg, ent->ce_location);
			continue;
		}
#ifdef LOC_MOVE
		/* save old location */
		if (ent->ce_flags & EF_IS_AT_LOCATION)
			ent->ce_flags |= EF_WAS_AT_LOCATION;
		else ent->ce_flags &= ~EF_WAS_AT_LOCATION;
#endif /* LOC_MOVE */
		ent->ce_flags &= ~EF_IS_AT_LOCATION;
		location = &location_head;
		while (location->next) {
			location = location->next;
			re_matches = re_exec(location->name);
			if (re_matches) {
				ent->ce_flags |= EF_IS_AT_LOCATION;
				/* printf("location flag set %s\n",ent->ce_location); */
				break;
			};
		}
	}

	/*
	 * Free locations.
	 */
	location = location_head.next;
	while (location) {
		next_location = location->next;
		if (location != &location_head)
			free(location);
		location = next_location;
	};

 done:
	(void)chdir(CRONTABDIR);
	(void)chdir(LIBDIR);
}
#endif /* LOC_CRON */
