#ifndef lint
static char Rcs_Id[] =
	"$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/runcmd.c,v 2.29 2000/04/13 03:19:36 johnh Exp $";
#endif

/*
 * $Log: runcmd.c,v $
 * Revision 2.29  2000/04/13 03:19:36  johnh
 * fixes for several bugs: loc-specific at jobs running multiple times and at-job reading, and geoffs mis-applied Y2K patch
 *
 * Revision 2.28  2000/02/08 08:56:42  geoff
 *
 * Integrate various changes from Geoff's branch into the now-master CVS
 * repository.  As part of this integration, I cleaned up some, but not
 * all, of the inconsistent coding style introduced by the LOC_CRON
 * changes.  Important and substantive changes include:
 *
 *     Configure	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     Makefile.SH	Tiny cleanup to an expr command.
 *     config.SH	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     errmsg.c	Add errlist_in_stdio support.
 *     mainloop.c	Use setjmp/longjmp to ensure that we are blown off the
 * 		select when we catch an alarm, even if we are running
 * 		on a system that has reliable signals.  Also add
 * 		reliable-signal support that properly resets the
 * 		signal mask in the longjmp case.
 *     parse.c	Fix a bug that could cause crontabs to be accepted
 * 		with the wrong UID.  Be a bit more careful about
 * 		parsing years.
 *     runcmd.c	Add setjmp/longjmp and reliable-signal support.  (See
 * 		mainloop.c).
 *     xcron.c	Use setitimer to make absolutely sure that all alarms
 * 		are disabled at startup.
 *
 * Revision 2.27  1999/07/27 01:22:00  johnh
 * location support for at jobs
 *
 * Revision 2.26  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.25  1998/12/17 23:19:07  geoff
 * Make sure at jobs only run once
 *
 * Revision 2.24  1998/07/30 20:42:28  geoff
 * Many changes, accumulated over far too much time: use a
 * non-restrictive umask to run cron jobs; support long names and Y2K
 * fixes; support mail-size limits.
 *
 * Revision 2.23  1995/11/27  01:59:47  geoff
 * Don't inform the user of at jobs that ran late due to overfull queues.
 * When reporting at-job errors, include the contents of the job.
 *
 * Revision 2.22  1995/05/11  05:05:51  geoff
 * Use setjmp/longjmp to handle reliable signals properly.  Add
 * MAILLOCKSLEEP as an ifdef.
 *
 * Revision 2.21  87/04/01  23:50:46  geoff
 * Set kidpid before using it (picky, picky).
 * 
 * Revision 2.20  87/03/31  22:34:07  geoff
 * Close the fifo before running child processes.  Feed the at job file to
 * the shell as standard input, rather than as an argument;  this solves
 * some protection problems.  Start at jobs running in the user's home
 * directory (usually they will then cd elsewhere).  Run the mailer with
 * execlp so shell scripts can be used.
 * 
 * Revision 2.19  87/01/10  19:20:21  geoff
 * Make a zero queue limit or process limit mean infinite (since it's
 * documented to work that way), and enable limits for root.  Make sure
 * a bunch of signals that cron ignores aren't ignored in the user process.
 * 
 * Revision 2.18  86/11/29  19:53:40  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.  Also add support for
 * systems without setpgrp(), and for Berkeley-style setpgrp() calls.
 * 
 * Revision 2.17  86/11/29  16:44:27  geoff
 * Fix a minor lint problem.
 * 
 * Revision 2.16  86/11/16  00:59:36  geoff
 * Add (untested) support for the BSD initgroups() call.
 * 
 * Revision 2.15  86/04/17  22:40:11  geoff
 * Put spawned commands in their own process group
 * 
 * Revision 2.14  86/04/02  00:03:02  geoff
 * Remove the third parameter to mailto, and modify it to keep a static
 * variable that permits only one mailing per ten seconds.
 * 
 * Revision 2.13  86/04/01  22:03:40  geoff
 * Move a whole bunch of code into mainloop.c.  This also involves a complete
 * rewrite of the routine 'cron'.
 * 
 * Revision 2.12  86/04/01  19:03:45  geoff
 * Make catchup return the system time.  Make nextwork a local variable to
 * cron().  Move some stuff into subroutines.
 * 
 * Revision 2.11  86/03/31  22:59:32  geoff
 * Remove the signal-based dumping stuff.  Add support for queue length
 * limits.
 * 
 * Revision 2.10  86/03/30  15:42:21  geoff
 * Switch debugging output to the DBPRINTn macros when possible
 * 
 * Revision 2.9  86/03/30  14:34:37  geoff
 * Major changes to take advantage of the sorted entry list.  Cron now sleeps
 * until the next time it has work, rather than waking up every minute and
 * searching for work.  It also aborts the work search as soon as it gets
 * past entries that are due.  Add temp-file cleanup, support for commands
 * from tellcron, and an exit message.
 * 
 * Revision 2.8  86/03/28  14:42:59  geoff
 * Add support for on-demand table dumps.  Modify findandrun to update
 * the ce_rtime field and re-insert the entry in its new sorted position
 * in the list.  Add code to nice user commands.  Fix some lint complaints.
 * 
 * Revision 2.7  86/03/25  19:57:12  geoff
 * Add support for at jobs.  Also add MAILLOCKSLEEP support.
 * 
 * Revision 2.6  86/01/04  16:17:42  geoff
 * Modify runcommand to queue up commands instead of just running them, and
 * add 'runqueues' to run the queues, giving at least some consideration to the
 * system load.  Change message formats to be closer to V.2 and to give
 * more useful information.  Change the interval of certain debugging output
 * during catchup so that we don't spend forever catching up at -x15.  Add
 * vfork support.  Send mail from the DAEMON uid/gid.
 * 
 * Revision 2.5  85/12/21  15:16:20  geoff
 * Make the cron time log file if it doesn't already exist.
 * 
 * Revision 2.4  85/11/25  16:53:26  geoff
 * Add code to support the year field.  Fix some lint complaints.
 * Fix an off-by-one bug in month-day processing.  Fix waitfor () to always do
 * a wait call, so we pick up defunct mail processes.
 * 
 * Revision 2.3  85/11/24  16:14:34  geoff
 * Major revisions to support getting information via the FIFO:  add
 * fifocatch, checkfifo, and fifogetc.  Remove newfiles and support for
 * the EF_NEWFILES bit, the files_changed variable, and automatic checking
 * for changed files.  Fix some lint complaints.
 * 
 * Revision 2.2  85/11/05  00:37:19  geoff
 * Move childproc structure to the header file, and bring up to date with
 * new flag names and garbage collection methods.
 * 
 * Revision 2.1  85/11/03  00:10:45  geoff
 * Major changes to support System V.2-style cron:  childproc structure,
 * passing of a controlled environment, and the routines "waitfor", "finishup",
 * and "mailto".  Stdout and stderr for each command are collected and mailed
 * to the owner of the crontab.
 * 
 * Revision 1.1  85/10/12  01:33:40  geoff
 * Break it up into separate modules
 * 
 */

#include "at.h"
#include "cron.h"
#include <errno.h>
#include <pwd.h>
#include <setjmp.h>
#include <stdio.h>

time_t		findandrun ();		/* Find runnable cmds and run them */
static void	queuecommand ();	/* Queue a command for later running */
static void	runcommand ();		/* Run one command */
static struct childproc *
		fexec ();		/* Fork and exec a command */
void		runqueues ();		/* Run the commands on the queues */
static FILE *	makefile ();		/* Create a temp file for process */
int		waitfor ();		/* Wait for a process to finish */
static void	finishup ();		/* Finish up after a child runs */
void		mailto ();		/* Send mail to a user */

int		alarm_caught = 0;	/* NZ if an alarm has been caught */
jmp_buf		alarm_jb;		/* Jump buffer for catching alarms */
int		alarm_jb_valid = 0;	/* NZ if alarm_jb has valid value */

extern unsigned	alarm ();		/* Set alarm clock timeout */
extern struct cronent enthead;		/* Head of cronent chain */
extern char *	errmsg ();		/* Return an error message string */
extern int	errno;			/* System error number */
extern void	_exit ();		/* Exit without buffer flushing */
extern int	fifofd;			/* Cron fifo's fd, so we can close it */
extern void	free ();		/* Return memory to the allocator */
extern void	insent ();		/* Insert entry into the entry list */
extern long	lseek ();		/* Seek to file position */
extern char *	malloc ();		/* Memory allocator */
extern time_t	nextrun ();		/* Figure out when entry runs next */
extern struct childproc
		prochead;		/* Head of child process list */
extern short	re_collect;		/* NZ if collection is needed later */
extern char *	realloc ();		/* Reallocate memory */
#ifdef DEBUG
extern int	reallyrun;		/* Never actually run commands if Z */
#endif
extern unsigned	sleep ();		/* Suspend process for a while */
extern time_t	start_timeout;		/* Timeout to apply during startup */
extern char *	strcpy ();		/* Copy strings */
extern time_t	syncinterval;		/* Interval between syncs, if any */
extern char *	tempfile ();		/* Create a temp file name */
extern time_t	time ();		/* Return system time */
extern char *	timenonl ();		/* Return a time w/o newline */
extern char *	tzenvvar;		/* Value of TZ environment variable */
extern struct cronuser userhead;	/* Head of cronuser chain */

time_t findandrun (when, flags, ignore)	/* Find and run commands */
    reg5 time_t		when;		/* When commands are to be run */
    reg4 int		flags;		/* Flags affecting our operation */
					/* If NZ, selects some commands */
					/* Zero means run all eligible */
    reg6 int		ignore;		/* Ignore entries with these flags */
    {
    reg1 struct cronent * entp;		/* Current entry in command list */
    reg2 struct cronent * nentp;	/* Next entry in command list */
    reg3 struct cronent * relink = NULL; /* List of commands we're relinking */

    ignore |= EF_DELETED;		/* We always ignore deleted entries */
    for (entp = enthead.ce_next;
      entp != &enthead
	&&  ((entp->ce_flags & ignore)  ||  entp->ce_rtime <= when);
      entp = nentp)
	{
	nentp = entp->ce_next;
	if ((entp->ce_flags & ignore) == 0)	/* One we need to process? */
	    {
	    if (flags)				/* Catchup mode? */
		{
		if ((entp->ce_flags & EF_HASRUN) == 0
		  &&  (flags & entp->ce_flags))	/* Right mode? */
		    {
		    runcommand (entp, entp->ce_flags, 0); /* Run right now */
		    if ((entp->ce_flags & EF_RUNALL) == 0)
			entp->ce_flags |= EF_HASRUN;
		    }
		}
	    else
		{
#ifdef LOC_CRON
#ifdef LOC_MOVE
		if (entp->ce_location == NULL  ||  loc_are_we_there_yet(entp))
		    {
		        queuecommand (entp); /* Queue command for execution */
		    }
#else /* ! LOC_MOVE */
		if (entp->ce_location == NULL
		  ||  (entp->ce_flags & EF_IS_AT_LOCATION))
		    queuecommand (entp); /* Queue command for execution */
#endif /* LOC_MOVE */
#else
		queuecommand (entp); /* Queue command for execution */
#endif		
		}
	    /*
	     * Requeue the command even if we didn't actually run it.
	     */
	    entp->ce_prev->ce_next = entp->ce_next; /* Delink it */
	    entp->ce_next->ce_prev = entp->ce_prev; /* ... */
	    entp->ce_prev = relink;	/* Save it on the relink list */
	    relink = entp;
	    }
	}
    /*
     * Re-insert the ones we ran into their new positions
     */
    for (entp = relink;  entp;  entp = nentp)
	{
	nentp = entp->ce_prev;
#ifdef LOC_CRON
	/* xxx: I don't remember why lcron is more agressive at running
	 * jobs here.  -johnh
	 * The old code used to not filter out atjobs here, but
	 * in nextrun.  Now nextrun returns all atjobs,
	 * so we filter here.
	 *
	 * But when we have LOC_ATLOC we have to let location-denied
	 * at jobs through.
	 */
	entp->ce_rtime =
		((entp->ce_flags & EF_ATJOB) && !entp->ce_location) ? 0
		: nextrun (when, entp);
#else /* ! LOC_CRON */
	entp->ce_rtime =
	  (entp->ce_flags & EF_ATJOB) ? 0 : nextrun (when, entp);
#endif /* LOC_CRON */
	if (entp->ce_rtime == 0)
	    {
	    entp->ce_flags |= EF_DELETED; /* Get rid of obsolete entries */
	    re_collect = 1;		/* Force a garbage collection */
	    }
	insent (entp);
	}
    /*
     * Finally, figure out what time the next command is due and return
     * that as a value.  This requires a loop only because, during catchup,
     * there are entries on the head of the command list that are not
     * runnable due to their flag values.
     *
     * If a sync interval is specified, we will return that instead of
     * the next run time if the next run time is past the sync interval.
     */
    if (syncinterval  &&  (flags == 0))
	when += syncinterval - when % syncinterval;
    else
	when = HUGETIME;
    for (entp = enthead.ce_next;  entp != &enthead;  entp = entp->ce_next)
	{
#ifdef LOC_CRON
	if (entp->ce_location == NULL  ||  loc_are_we_there_yet (entp))
	    {
#endif
	    if (entp->ce_rtime > when)	/* Past next sync interval? */
		return when;		/* If so, just return that time */
	    else if ((entp->ce_flags & ignore) == 0)
		return entp->ce_rtime;
#ifdef LOC_CRON	   
	    }
#endif
	}
    /*
     * There is no command due because the entry queue is empty.  Return
     * the next sync time (HUGETIME if there is no sync interval).
     */
    return when;
    }

static void queuecommand (entp)		/* Queue a command for running later */
    reg3 struct cronent * entp;		/* Entry describing the command */
    {
    reg1 struct cmdqueue * queueent;	/* New command queue entry */
    reg2 struct cronuser * user;	/* User running this command */

#ifdef LOC_CRON
    /*
     * This is kind of hacky...
     * for some reason, location-specific at jobs get run
     * a bunch of times.  To avoid this, we set a flag
     * that the job has fired when it's queued here,
     * and we ignore jobs that have fired and are in the queue.
     * When the job completes, it's taken off the queue.
     * (Mabye Geoff has a more elegant solution :-( ).  -johnh, 12-Apr-00
     */
    if (entp->ce_location && (entp->ce_flags & EF_ATJOB))
        entp->ce_flags |= EF_ONLY_ON_MOVE_FIRED;
#endif /* LOC_CRON */
    user = entp->ce_file->cf_user;
    if (user->cu_qlim > 0  &&  user->cu_qcount >= user->cu_qlim)
	{
	DBPRINT2 (4,
	  "queuecommand:  discarding %s (for %s) -- queue too long\n",
	  entp->ce_command ? entp->ce_command : "at job",
	  user->cu_name + EL_LOGNAME);
	user->cu_flags |= UF_LOSTJOBS;
	return;
	}
    user->cu_qcount++;
    queueent = (struct cmdqueue *) malloc (sizeof (struct cmdqueue));
    if (queueent == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "queuecommand:  out of memory for %s\n",
	  entp->ce_command ? entp->ce_command : "at job");
	return;
	}
    queueent->cq_prev = user->cu_queue.cq_prev;
    queueent->cq_next = &(user->cu_queue);
    queueent->cq_prev->cq_next = queueent;
    user->cu_queue.cq_prev = queueent;
    queueent->cq_ent = entp;
    queueent->cq_flags = 0;
    }

static void runcommand (entp, how, qflags) /* Run a command */
    reg2 struct cronent *	entp;	/* Entry describing the command */
    reg3 int			how;	/* How to run it (flags) */
    int				qflags;	/* Flags from queuing it */
    {
    reg1 struct childproc *	child;	/* Proc desriptor for spawned proc */
    reg4 int			kidpid;	/* Process ID of spawned child */

#ifdef LOC_CRON
    if (entp->ce_location && !(entp->ce_flags & EF_IS_AT_LOCATION))
	return;
#ifdef LOC_MOVE
    /* check edge triggering */
    if (entp->ce_location && (entp->ce_flags & EF_ONLY_ON_MOVE))
	{
	if (entp->ce_flags & EF_WAS_AT_LOCATION)
	    {
	    DBPRINT0(15, "runcommand: returning because of failed edge detection.\n");
	    return;
	    }
	/* indicate that the command has run at the new location */
	entp->ce_flags |= EF_WAS_AT_LOCATION;
	}
#endif /* LOC_MOVE */
#endif
    child = fexec (entp, qflags);	/* Fork the command */
    if (child == NULL)
	{
	/* another place where I had a failure ---johnh */
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "runcommand:  fexec failed on %s.\n", entp->ce_file->cf_name);
	return;
	}
    kidpid = child->cp_pid;
    (void) fseek (stderr, 0L, 2);	/* Make sure we're at EOF */
    if (entp->ce_flags & EF_ATJOB)	/* Log the command */
	(void) fprintf (stderr, ">  %s %d %s %s < %s/%s\n",
	  entp->ce_file->cf_user->cu_name + EL_LOGNAME, kidpid,
	  timenonl (time ((long *) NULL)), ATSHELL, ATJOBDIR,
	  entp->ce_file->cf_name);
    else
	(void) fprintf (stderr, ">  %s %d %s %s\n",
	  entp->ce_file->cf_user->cu_name + EL_LOGNAME, kidpid,
	  timenonl (time ((long *) NULL)), entp->ce_command);
    /*
     * The command is running.  If we are running in one of the sequential
     * catchup modes, we must wait for it.  Otherwise, we must link it
     * onto the child-process list for later waits.
     */
    if (child  &&  (how & (EF_WAIT | EF_SEQSPAWN))) /* Do we need to wait? */
	{
	/*
	 * Wait for the child to terminate.  Waitfor returns a value
	 * greater than zero if the wait failed but can be retried.
	 */
	while (waitfor (kidpid) > 0)
	    ;				/* Wait for child */
	/* Waitfor takes care of deallocating and delinking child */
	}
    }

/*
 * Fork and execute a child process.  A childproc structure is allocated
 * and linked into the child process list to describe the child.  The
 * return value is a pointer to the child process structure.
 */
static struct childproc * fexec (entp, qflags) /* Fork and exec a command */
    reg2 struct cronent * entp;		/* Pointer to cronent to run */
    int			qflags;		/* Flags from queuing the command */
    {
    reg1 struct childproc * child;	/* Process struct to describe kid */
    reg6 FILE *		errfile;	/* File descriptor of stderr */
    reg7 FILE *		inpfile;	/* File descriptor of input file */
    reg3 int		kidpid;		/* Process ID of child */
    reg8 FILE *		outfile;	/* File descriptor of stdout file */
#ifdef IGROUPS
    reg5 struct passwd * pw;		/* Password entry for user */
#endif
    int			saveerr;	/* Place to save errno b4 clobber */
    reg4 struct cronuser * userp;	/* User running this job */
    static char *	newenv[] =	/* New environment for child */
	{
	DEFPATH, DEFSHELL,
	NULL,				/* Space for HOME */
	NULL,				/* Space for LOGNAME */
	NULL,				/* Space for TZ, if available */
	NULL				/* Terminator */
	};
    
    child = (struct childproc *) malloc (sizeof (struct childproc));
    if (child == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "fexec:  out of memory for process descriptor\n");
	return NULL;
	}
    child->cp_ent = entp;
    (void) strcpy (child->cp_fbase, TEMPBASE);	/* Set up base of names */
    (void) tempfile (child->cp_fbase);		/* ... */
    outfile = makefile (entp, child->cp_fbase, 'o', &child->cp_oprefix);
    DBPRINT2 (13, "fexec (%s): file base is %s\n",
      entp->ce_command ? entp->ce_command : "at job", child->cp_fbase);
    if (outfile == NULL)
	{
	free (child);
	return NULL;
	}
    errfile = makefile (entp, child->cp_fbase, 'e', &child->cp_eprefix);
    if (errfile == NULL)
	{
	child->cp_fbase[TEMPTYPELOC] = 'o';
	(void) unlink (child->cp_fbase);
	free (child);
	return NULL;
	}
    userp = entp->ce_file->cf_user;
    /*
     * Let the user know about jobs that were delayed due to queueing.
     * However, at John Heidemann's request, don't make this complaint
     * for at jobs.  The theory is that people are more likely to
     * schedule multiple at jobs (e.g., reminders) for the same time,
     * and also that people mind less when at jobs are delayed a bit.
     * This theory may or may not be valid.
     */
    if ((qflags & QF_RANLATE) != 0  &&  (entp->ce_flags & EF_ATJOB) == 0)
	{
	(void) fprintf (errfile,
	  "cron:  this job was run later than the scheduled time because\n");
	(void) fprintf (errfile,
	  "       you already had %d outstanding processes.\n\n",
	  userp->cu_proclim);
	}
    if (userp->cu_flags & UF_LOSTJOBS)
	{
	(void) fprintf (errfile,
	  "cron:  some of your requested cron or at jobs (not\n");
	(void) fprintf (errfile,
	  "       this one) were discarded because you exceeded\n");
	(void) fprintf (errfile,
	  "       your queueing limit of %d\n\n", userp->cu_qlim);
	userp->cu_flags &= ~UF_LOSTJOBS;
	}
    if (entp->ce_input)
	{
	/* Create an input file.  Note that makefile changes cp_fbase. */
	inpfile = makefile (entp, child->cp_fbase, 'i', (short *) NULL);
	if (inpfile == NULL)
	    {
	    child->cp_fbase[TEMPTYPELOC] = 'e';
	    (void) unlink (child->cp_fbase);
	    child->cp_fbase[TEMPTYPELOC] = 'o';
	    (void) unlink (child->cp_fbase);
	    free (child);
	    return NULL;
	    }
	(void) fwrite (entp->ce_input, 1, strlen (entp->ce_input), inpfile);
					/* Write the input file */
	(void) fwrite ("\n", 1, 1, inpfile);
					/* Follow it with a final newline */
	(void) fclose (inpfile);
	inpfile = fopen (child->cp_fbase, "r"); /* Open the file for reading */
	if (inpfile == NULL)
	    {
	    saveerr = errno;
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "fexec:  couldn't reopen input file %s (%s) for %s\n",
	      child->cp_fbase, errmsg (saveerr),
	      entp->ce_command ? entp->ce_command : "at job");
	    (void) unlink (child->cp_fbase);
	    child->cp_fbase[TEMPTYPELOC] = 'e';
	    (void) unlink (child->cp_fbase);
	    child->cp_fbase[TEMPTYPELOC] = 'o';
	    (void) unlink (child->cp_fbase);
	    free (child);
	    return NULL;
	    }
	(void) unlink (child->cp_fbase); /* We can safely unlink now */
	}
    else if (entp->ce_flags & EF_ATJOB)
	{
	(void) chdir (ATJOBDIR);
	inpfile = fopen (entp->ce_file->cf_name, "r");
	/*
	 * If the input file wouldn't open, we put out a message, but
	 * otherwise continue;  this is simpler than cleaning up.
	 * The shell will be fed /dev/null as input, which is fine.
	 */
	if (inpfile == NULL)
	    (void) fprintf (errfile,
	      "cron:  couldn't open at job %s for input\n",
	      entp->ce_file->cf_name);
	(void) chdir (LIBDIR);
	}
    else
	inpfile = NULL;
    (void) fflush (stdout);
    (void) fflush (stderr);
#ifdef IGROUPS
    /*
     * Read the user's password entry so we can do initgroups call.  This
     * must be done before the vfork call, because it uses stdio.
     */
    pw = getpwuid (entp->ce_uid);
#endif
    /*
     * Finish preparing the environment.  newenv[0] and newenv[1] have
     * already been set to their proper values.
     */
    newenv[2] = userp->cu_home;
    newenv[3] = userp->cu_name;
    newenv[4] = tzenvvar;
    while ((kidpid = vfork ()) == -1  &&  errno == EAGAIN)
	(void) sleep (5);
    if (kidpid < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "cron:  couldn't fork:  %s\n",
	  errmsg (saveerr));
	if (inpfile != NULL)
	    (void) fclose (inpfile);
	child->cp_fbase[TEMPTYPELOC] = 'e';
	(void) unlink (child->cp_fbase);
	child->cp_fbase[TEMPTYPELOC] = 'o';
	(void) unlink (child->cp_fbase);
	(void) fclose (errfile);
	(void) fclose (outfile);
	free (child);
	return NULL;
	}
    else if (kidpid != 0)		/* Parent returns to caller now */
	{
	if (inpfile != NULL)
	    (void) fclose (inpfile);
	(void) fclose (errfile);
	(void) fclose (outfile);
	child->cp_pid = kidpid;
	prochead.cp_prev->cp_next = child;
	child->cp_next = &prochead;
	child->cp_prev = prochead.cp_prev;
	prochead.cp_prev = child;
	userp->cu_procs++;		/* Count this process */
	return child;
	}
    else				/* Child runs the command */
	{
	(void) nice (userp->cu_nice);	/* Adjust our priority */
#ifdef HASSETPGRP
#ifdef SETPGRPARGS
	(void) setpgrp (0, getpid ());	/* Get out of cron's process group */
#else
	(void) setpgrp ();		/* Get out of cron's process group */
#endif
#endif
	if (inpfile != NULL)
	    {
	    (void) close (fileno (stdin)); /* Get rid of stdin */
	    (void) dup (fileno (inpfile)); /* Provide temp file as stdin */
	    (void) close (fileno (inpfile)); /* Get rid of temp file fd */
	    }
	/*
	 * Thanks to the way Berkeley implemented vfork, we must avoid
	 * using stdio from here on out.  That's actually not so bad,
	 * since it saves us from doing interesting fdopen's to make stuff
	 * work right.
	 */
	(void) close (fileno (stdout));	/* Set up stdout and stderr */
	(void) dup (fileno (outfile));
	(void) close (fileno (outfile));
	(void) close (fileno (stderr));
	(void) dup (fileno (errfile));
	(void) close (fileno (errfile));
	/*
	 * Use a fairly non-restrictive umask to run cron jobs.  This
	 * is a bit of a security risk, but it's historically
	 * compatible.
	 */
	(void) umask (JOBUMASK);
#ifdef IGROUPS
	/*
	 * Initialize the process's groups, using the password entry for the
	 * uid's base group.  This protects us in case the at file gid
	 * doesn't match the base gid.
	 */
	initgroups (pw->pw_name, pw->pw_gid);
#endif
	(void) setgid (entp->ce_gid);	/* Use user's real id's for chdir! */
	(void) setuid (entp->ce_uid);
	if (chdir (userp->cu_home + EL_HOME) < 0)
	    {
	    (void) lseek (2, 0L, 2);
#define ERR_PART1 "cron:  couldn't cd to "
#define ERR_PART2 ";  using "
	    (void) write (2, ERR_PART1, sizeof ERR_PART1 - 1);
	    (void) write (2, userp->cu_home + EL_HOME,
	      (unsigned) strlen (userp->cu_home + EL_HOME));
	    (void) write (2, ERR_PART2, sizeof ERR_PART2 - 1);
	    (void) write (2, SAFEHOME, sizeof SAFEHOME - 1);
	    (void) write (2, "\n", 1);
#undef ERR_PART1
#undef ERR_PART2
	    (void) chdir (SAFEHOME);
	    }
	(void) signal (SIGINT, SIG_DFL); /* Re-default signals we ignored */
	(void) signal (SIGHUP, SIG_DFL);
	(void) signal (SIGQUIT, SIG_DFL);
	(void) signal (SIGTERM, SIG_DFL); /* I don't know how this got */
					/* ..ignored, but it did */
	close (fifofd);			/* Close the cron fifo */
	if (start_timeout)		/* Set up timeout if requested */
	    (void) alarm ((unsigned) start_timeout);
	/* Run the command */
#ifdef DEBUG
	if (reallyrun)
	    {
	    if (entp->ce_flags & EF_ATJOB)
		(void) execle (ATSHELL, ATARGV0, (char *) NULL, newenv);
	    else
		{
		(void) execle (CMDSHELL, CMDARGV0,
		  "-c", entp->ce_command, (char *) NULL,
		  newenv);
		}
	    }
#else
	if (entp->ce_flags & EF_ATJOB)
	    (void) execle (ATSHELL, ATARGV0, (char *) NULL, newenv);
	else
	    {
	    (void) execle (CMDSHELL, CMDARGV0,
	      "-c", entp->ce_command, (char *) NULL,
	      newenv);
	    }
#endif
	_exit (errno ? errno : -1);	/* Here only on failures */
	/* NOTREACHED */
	}
    }

void runqueues ()			/* Run the commands on the queues */
    {
    reg1 struct cmdqueue * ent;		/* Queue entry we're running */
    reg2 struct cmdqueue * nent;	/* Next queue entry */
    reg3 struct cronuser * user;	/* User whose queues we're checking */

    for (user = userhead.cu_next;  user != &userhead;  user = user->cu_next)
	{
	for (ent = user->cu_queue.cq_next;
	  ent != &(user->cu_queue);
	  ent = nent)
	    {
	    nent = ent->cq_next;
	    if (user->cu_proclim > 0  &&  user->cu_procs >= user->cu_proclim)
		{
		ent->cq_flags |= QF_RANLATE;
		DBPRINT2 (4,
		  "runqueues:  skipping %s (for %s) -- too many procs\n",
		  ent->cq_ent->ce_command
		    ? ent->cq_ent->ce_command : "at job",
		  user->cu_name + EL_LOGNAME);
		}
	    else
		{
		user->cu_qcount--;
		runcommand (ent->cq_ent, 0, ent->cq_flags);
		ent->cq_prev->cq_next = nent;
		nent->cq_prev = ent->cq_prev;
		free (ent);
		}
	    }
	}
    }

/* Create a temp file for a process.  Output files have a header written
 * into them, and the header size is returned in the size parameter.
 */
static FILE * makefile (entp, basename, code, sizeptr)
    reg1 struct cronent * entp;		/* Crontab entry for process */
    reg2 char *		basename;	/* Base name for files, MODIFIED */
    reg6 int		code;		/* Code for the file (i, o, or e) */
    reg7 short *	sizeptr;	/* Pointer to where to put size */
    {
    reg9 FILE *		atjobfile;	/* For reading at job's contents */
    char		buf[512];	/* Buffer for listing at job */
    reg5 int		fd;		/* Descriptor for new file */
    reg8 int		len;		/* Length of data in buf */
    reg4 char *		namep;		/* Pointer to filename for at jobs */
    int			saveerr;	/* Place to save errno b4 clobber */
    struct stat		statbuf;	/* Stat buffer for getting size */
    reg3 FILE *		stdfile;	/* Access to file through stdio */
    int			year;		/* Year that an 'at' job was to run */

    DBPRINT4 (14, "makefile (0x%x, %s, '%c', 0x%x\n", entp, basename,
      code, sizeptr);
    basename[TEMPTYPELOC] = code;	/* Put file code in base name */
    if ((fd = creat (basename, 0600)) < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "makefile:  couldn't create standard %s file %s (%s) for %s\n",
	  (code == 'i') ? "input" : ((code == 'o') ? "output" : "error"),
	  basename, errmsg (saveerr),
	  entp->ce_command ? entp->ce_command : "at job");
	return NULL;
	}
    stdfile = fdopen (fd, "w");
    if (stdfile == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "makefile:  couldn't fdopen %s after creating it\n",
	  basename);
	(void) close (fd);
	return NULL;
	}
    if (code == 'e'  ||  code == 'o')	/* Does file need an intro header? */
	{
	(void) fprintf (stdfile, "To: %s\n",
	  entp->ce_file->cf_user->cu_name + EL_LOGNAME);
	if (entp->ce_flags & EF_ATJOB)
	    {
	    (void) fprintf (stdfile,
	      "Subject: at %soutput\n\n", code == 'e' ? "error " : "");
	    namep = entp->ce_file->cf_name;
	    (void) fprintf (stdfile, "Your at job scheduled for ");
#ifdef LONGNAMES
	    if (namep[WHEN_FLAG] == WHEN_FLAG_VALUE)
		(void) fprintf (stdfile, "%4.4s/%2.2s/%2.2s at %2.2s:%2.2s ",
		  namep + WHEN_YEAR, namep + WHEN_MONTH, namep + WHEN_DAY,
		  namep + WHEN_HOUR, namep + WHEN_MINUTE);
	    else
		{
		(void) sscanf(&namep[WHEN_OLD_YEAR], "%2.2d", &year);
		if (year < Y2K_THRESHOLD)
		    year += Y2K_2000;
		else
		    year += Y2K_1900;
		(void) fprintf (stdfile, "%4.4d/%2.2s/%2.2s at %2.2s:%2.2s ",
		  year, namep + WHEN_OLD_MONTH, namep + WHEN_OLD_DAY,
		  namep + WHEN_OLD_HOUR, namep + WHEN_OLD_MINUTE);
		}
#else /* LONGNAMES */
	    (void) sscanf(&namep[WHEN_YEAR], "%2.2d", &year);
	    if (year < Y2K_THRESHOLD)
		year += Y2K_2000;
	    else
		year += Y2K_1900;
	    (void) fprintf (stdfile, "%4.4d/%2.2s/%2.2s at %2.2s:%2.2s ",
	      year, namep + WHEN_MONTH, namep + WHEN_DAY, namep + WHEN_HOUR,
	      namep + WHEN_MINUTE);
#endif /* LONGNAMES */
	    (void) fprintf (stdfile, "produced %s%s",
	      code == 'e' ? "error " : "",
	      "output,\nwhich follows the job listing below:\n\n");
	    (void) fprintf (stdfile,
	      "--------------------at job--------------------\n");
	    (void) chdir (ATJOBDIR);
	    atjobfile = fopen (entp->ce_file->cf_name, "r");
	    if (atjobfile == NULL)
		{
		/*
		 * A strange error, but this actually
		 * happened to me. --johnh, 11-Jul-99
		 */
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr,
				"makefile:  couldn't fopen %s for reading\n",
				entp->ce_file->cf_name);
		(void) fclose(stdfile);
		return NULL;
		}
	    (void) chdir (LIBDIR);
	    /*
	     * Look through the file trying to find the line
	     * $shell << 'UNIQUESTR'.  If we find it, print
	     * out the following lines until we reach UNIQUESTR.
	     * These are presumably the commands in the job.
	     */
	    while (fgets (buf, sizeof buf, atjobfile) != NULL)
		{
		len = strlen (buf);
		if (buf[len - 1] != '\n')
		    {
		    while ((len = getc (atjobfile)) != '\n'
		      &&  len != EOF)
			;
		    }
		else if (len > sizeof UNIQUESTR + 2
		  &&  strncmp (&buf[len - sizeof UNIQUESTR - 1],
		    UNIQUESTR, sizeof UNIQUESTR - 1) == 0)
		    {
		    while (fgets (buf, sizeof buf, atjobfile) != NULL)
			{
			if (strncmp (buf, UNIQUESTR,
			    sizeof UNIQUESTR - 1)
			  == 0
			  &&  buf[sizeof UNIQUESTR - 1] == '\n')
			    break;	/* End of commands */
			(void) fputs (buf, stdfile);
			}
		    break;
		    }
		}
	    (void) fclose (atjobfile);
	    (void) fprintf (stdfile,
	      "--------------------%soutput--------------------\n",
	      code == 'e' ? "error " : "");
	    }
	else
	    {
	    (void) fprintf (stdfile,
	      "Subject: cron %soutput\n\n", code == 'e' ? "error " : "");
	    (void) fprintf (stdfile,
	      "Your cron command \"%s\" produced %soutput, which follows:\n\n",
	      entp->ce_command, code == 'e' ? "error " : "");
	    }
	(void) fflush (stdfile);
	if (fstat (fileno (stdfile), &statbuf) < 0)
	    *sizeptr = 0;		/* This will always force mail */
	*sizeptr = statbuf.st_size;
	}
    return stdfile;
    }

/*
 * Wait for a child to finish.  If pid is zero, wait for any child.
 * If a pid is specified, returns only when that pid has disappeared from
 * the child process list.  If no pid is specified, waitfor returns as
 * soon as the wait () call returns, regardless of the reason.
 *
 * Waitfor is also responsible for post-processing (removing temporary files
 * and mailing them to the user).
 *
 * Return value is zero if a successful wait occurred, positive if the wait
 * failed but children remain (e.g., alarms), and negative if the wait
 * failed because no children remain.
 */
int waitfor (pid)
    reg3 int			pid;	/* Process ID to wait for */
    {
    reg1 struct childproc *	child;	/* Child process descriptor to match */
    reg2 int			deadpid; /* PID of last process to die */
#ifdef RELIABLE_SIGNALS
    int				oldsigs; /* Previous signal mask */
#endif /* RELIABLE_SIGNALS */
    reg4 int			saveerr; /* Space to save errno */
    int				status;	/* Status of last process to die */
#ifdef HASSIGPROCMASK
    sigset_t			unblockset; /* Signals to be unblocked */
#endif /* HASSIGPROCMASK */

    DBPRINT1 (13, "waitfor (%d)\n", pid);
    if (pid != 0)			/* Looking for a particular process? */
	{
	for (child = prochead.cp_next;	/* Yes, see if it died earlier */
	  child != &prochead;
	  child = child->cp_next)
	    {
	    DBPRINT4 (15, "PID %d o/e size %d/%d base %s\n",
	      child->cp_pid, child->cp_oprefix, child->cp_eprefix,
	      child->cp_fbase);
	    if (child->cp_pid == pid)
		break;
	    }
	if (child == &prochead)		/* Did we find a child descriptor? */
	    {
	    DBPRINT0 (14, "waitfor:  child must have died earlier\n");
	    return 0;			/* No, must have terminated earlier */
	    }
	}
    do
	{
#ifdef RELIABLE_SIGNALS
	oldsigs = siggetmask ();
#endif /* RELIABLE_SIGNALS */
	if (setjmp (alarm_jb))
	    {
	    /* An alarm has been caught;  return to the caller. */
	    alarm_jb_valid = 0;
	    DBPRINT0 (15, "waitfor:  returning due to newly-caught alarm\n");
	    alarm_caught = 0;
#ifdef HASSIGPROCMASK
	    /*
	     * Since longjmp didn't return properly from the signal
	     * handler, a POSIX system will wind up with the alarm
	     * blocked.  So we have to unblock the signal again.
	     */
	    sigemptyset (&unblockset);
	    sigaddset (&unblockset, SIGALRM);
	    sigprocmask (SIG_UNBLOCK, &unblockset, (sigset_t *) NULL);
#endif /* HASSIGPROCMASK */
#ifdef RELIABLE_SIGNALS
	    /*
	     * Since longjmp didn't return properly from the signal
	     * handler, a BSD system will wind up with the alarm
	     * blocked.  So we have to restore the old mask, which we
	     * have carefully saved, to unblock it.
	     */
	    sigsetmask (oldsigs);
#endif /* RELIABLE_SIGNALS */
	    return 1;
	    }
	alarm_jb_valid = 1;
	if (alarm_caught)
	    {
	    /* An alarm has been caught;  return to the caller. */
	    alarm_jb_valid = 0;
	    DBPRINT0 (15,
	      "waitfor:  returning due to previously-caught alarm\n");
	    alarm_caught = 0;
	    return 1;
	    }
	deadpid = wait (&status);	/* Wait for somebody to die */
	alarm_jb_valid = 0;
	if (deadpid < 0)
	    {
	    saveerr = errno;
	    DBPRINT1 (15, "waitfor:  returning due to error: %s\n",
	      errmsg (saveerr));
	    alarm_caught = 0;
	    errno = saveerr;
	    return (saveerr == ECHILD) ? -1 : 1;
	    }
	DBPRINT1 (15, "waitfor:  child %d died\n", deadpid);
	for (child = prochead.cp_next;
	  child != &prochead;
	  child = child->cp_next)
	    {
	    if (child->cp_pid == deadpid)
		{
		finishup (child, status); /* This unlinks the child struct */
		break;
		}
	    }
	}
    while (pid != 0  &&  deadpid != pid);
    return 0;
    }

/* ARGSUSED */
static void finishup (child, status)	/* Finish up after a child runs */
    reg2 struct childproc *	child;	/* Descriptor of dead child process */
    int				status;	/* It's return status */
    {
    int				saveerr; /* Place to save errno b4 clobber */
    struct stat			statbuf; /* Buffer for file status */
    reg1 struct cronuser *	user;	/* User who ran this child */

    user = child->cp_ent->ce_file->cf_user;
    (void) fprintf (stderr, "<  %s %d %s - %d/%d\n",
      user->cu_name + EL_LOGNAME, child->cp_pid,
      timenonl (time ((long *) NULL)),
      (status >> 8) & 0xFF, status & 0xFF);
    /*
     * If the stdout file is of nonzero size, mail it to the user.
     * Then unlink it, even if the stat failed.
     */
    child->cp_fbase[TEMPTYPELOC] = 'o';
    if (stat (child->cp_fbase, &statbuf) >= 0
      &&  statbuf.st_size != child->cp_oprefix)
	mailto (user, child->cp_fbase);
    (void) unlink (child->cp_fbase);
    /*
     * Repeat for the standard-error file
     */
    child->cp_fbase[TEMPTYPELOC] = 'e';
    if (stat (child->cp_fbase, &statbuf) >= 0
      &&  statbuf.st_size != child->cp_eprefix)
	mailto (user, child->cp_fbase, user);
    (void) unlink (child->cp_fbase);
    /*
     * Uncount this process in the user's process limit
     */
    if (user->cu_procs > 0)
	user->cu_procs--;
    /*
     * If the job was an 'at' job, delete its entry and file
     */
    if (child->cp_ent->ce_flags & EF_ATJOB)
	{
	child->cp_ent->ce_flags |= EF_DELETED;
	child->cp_ent->ce_file->cf_flags |= FF_DELETED;
	(void) chdir (ATJOBDIR);
	if (unlink (child->cp_ent->ce_file->cf_name) < 0)
	    {
	    saveerr = errno;
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr, "cron: couldn't unlink at job %s: %s\n",
	      child->cp_ent->ce_file->cf_name, errmsg (saveerr));
	    }
	(void) chdir (LIBDIR);
	re_collect = 1;			/* Force a garbage collection */
	}
    /*
     * We are now done. (The input file was unlinked long ago, and went away
     * when the child departed.)  Remove the child from the process list.
     */
    child->cp_prev->cp_next = child->cp_next;
    child->cp_next->cp_prev = child->cp_prev;
    free (child);
    /*
     * Since this child may have been pointing to an entry which is no
     * longer needed, force a garbage collection.
     */
    if ((child->cp_ent->ce_flags & EF_DELETED)
      ||  (child->cp_ent->ce_file->cf_flags & FF_DELETED)
      ||  (child->cp_ent->ce_file->cf_user->cu_flags & UF_MARKDEL))
	re_collect = 1;
    }

void mailto (user, file)		/* Send mail to a user */
    reg2 struct cronuser * user;	/* User structure for recipient */
    reg3 char *		file;		/* Name of the file to mail */
    {
#ifdef MAILLOCKSLEEP
    reg5 time_t		curnow;		/* Current time of day */
#endif /* MAILLLOCKSLEEP */
    reg1 int		fd;		/* File descriptor for file to mail */
    reg4 int		kidpid;		/* Process ID of child */
#ifdef MAILLOCKSLEEP
    static time_t	lastmail = 0;	/* Time last mail was sent */
#endif /* MAILLLOCKSLEEP */
    struct stat		mailstat;	/* Buffer for mail-file status */
    int			saveerr;	/* Place to save errno b4 clobber */

#ifdef MAILLOCKSLEEP
    lastmail += MAILLOCKSLEEP;		/* Figure when it will be ok to mail */
    curnow = time ((long *) NULL);	/* See what time it is now */
    if (curnow > lastmail)		/* Past legal mailing time? */
	lastmail = curnow;		/* ..if so, reset legal time */
#endif /* MAILLLOCKSLEEP */
    DBPRINT2 (14, "mailto (%s, %s)\n", user->cu_name + EL_LOGNAME, file);
    /*
     * Note that there's a race condition here: if somebody replaces
     * the temp file with an appropriate symlink, they can trick us
     * into mailing them an arbitrary file.  We depend on protection
     * of the temporary directory to prevent this.
     */
    fd = open (file, 2);		/* Open the file */
    if (fd < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "cron:  couldn't open %s to mail it to %s: %s\n",
	  file, user, errmsg (saveerr));
	return;
	}
    if (fstat (fd, &mailstat) >= 0)
	{
	/*
	 * Limit the size of the mail file if necessary
	 */
	if (user->cu_mailsizelim != 0
	  &&  user->cu_mailsizelim < mailstat.st_size)
	    {
#ifdef HASFTRUNCATE
	    (void) ftruncate (fd, user->cu_mailsizelim);
#else /* HASFTRUNCATE */
#ifdef HASTRUNCATE
	    (void) truncate (file, user->cu_mailsizelim);
#else /* HASTRUNCATE */
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "cron:  unable to truncate %s to %d bytes (no system call)\n",
	      file, user->cu_mailsizelim);
#endif /* HASTRUNCATE */
#endif /* HASFTRUNCATE */
	    }
	}
    while ((kidpid = vfork ()) == -1  &&  errno == EAGAIN)
	(void) sleep (5);
    if (kidpid < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "cron:  couldn't fork to mail to %s:  %s\n",
	  user->cu_name + EL_LOGNAME, errmsg (saveerr));
	(void) close (fd);
	return;
	}
    else if (kidpid != 0)		/* Parent returns to caller now */
	{
	(void) close (fd);
	return;
	}
    else				/* Child runs the mailer */
	{
	(void) close (fileno (stdin));	/* Get rid of stdin */
	(void) dup (fd);		/* Provide mail file as stdin */
	(void) close (fd);		/* Get rid of mail file fd */
	(void) chdir (SAFEHOME);
	(void) setgid (DAEMONGID);	/* Send mail as 'daemon' */
	(void) setuid (DAEMONUID);	/* ... */
#ifdef MAILLOCKSLEEP
	/*
	 * Some mailers (e.g. V7, V.0) have a defective locking scheme that
	 * can produce mail errors if a program spawns mail twice quickly
	 * in succession.  To avoid this, this routine keeps track of when
	 * it last sent mail (to anybody) and sleeps to make sure that at
	 * most one mailer is spawned per MAILLOCKTIME.
	 */
	if (lastmail > curnow)
	    (void) sleep ((unsigned int) (lastmail - curnow));
#endif /* MAILLLOCKSLEEP */
	/* Run the mailer */
	(void) execlp (MAILER, MAILARGV0, user->cu_name + EL_LOGNAME,
	  (char *) NULL);
	_exit (errno ? errno : -1);	/* Here only on failures */
	/* NOTREACHED */
	}
    }
