#ifndef lint
static char Rcs_Id[] =
    "$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/sendcmd.c,v 2.9 1998/07/30 20:42:28 geoff Exp $";
#endif

/*
 * $Log: sendcmd.c,v $
 * Revision 2.9  1998/07/30 20:42:28  geoff
 * Add BSD socket support.
 *
 * Revision 2.8  1995/11/27  01:59:49  geoff
 * Keep the lock file in ASCII format for easy human reading and
 * modification.  If the time-log file hasn't been updated recently,
 * assume that cron isn't active.
 *
 * Revision 2.7  1995/05/11  05:05:53  geoff
 * Minor RCS header changes.
 *
 * Revision 2.6  87/01/10  19:20:59  geoff
 * Make the return codes symbolic.  Remove the code that made the fifo if
 * it didn't exist;  since a nonexistent fifo implies no cron is running,
 * there is not a lot of point in creating the thing.
 * 
 * Revision 2.5  86/11/29  19:54:18  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.4  86/04/17  22:00:44  geoff
 * Add a third argument to sendcommand to allow suppressing of the 'queued'
 * error message.
 * 
 * Revision 2.3  86/03/28  14:43:43  geoff
 * Fix a lint complaint.
 * 
 * Revision 2.2  86/03/25  19:58:04  geoff
 * Add setlibdir and getname, and do some improving.
 * 
 * Revision 2.1  86/01/19  16:04:21  geoff
 * Initial Revision (taken from an earlier version of crontab.c)
 * 
 */

#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include "cron.h"

int		sendcommand ();		/* Send command to the cron daemon */
static int	cronpid ();		/* Return PID of cron process */
void		setlibdir ();		/* Change directories to LIBDIR */
char *		getname ();		/* Get user's name */

struct passwd *	userpwent;		/* User's password entry */
					/* Set by getname */

extern void	endpwent ();		/* Close password file */
extern struct passwd * getpwnam ();	/* Get a password entry by name */
extern struct passwd * getpwuid ();	/* Get a password entry by uid */
extern char *	errmsg ();		/* Return system error message */
extern void	exit ();		/* Exit the program */
extern char *	malloc ();		/* Memory allocater */
extern char *	progname;		/* Name of program that called us */
extern char *	strcpy ();		/* String copy routine */

/*
 * Send a command to the cron daemon.  Uses the cron lock file to discover
 * the process ID of cron, and then sends the command over the FIFO.
 * Finally, it signals cron with a READFIFO signal.
 *
 * The caller of this routine must ensure that the current directory
 * is LIBDIR.
 *
 * Return value:  0 if the command was sent successfully, 1 if the daemon
 *   wasn't running, and 2 if there was a problem other than a non-running
 *   daemon.
 */
int sendcommand (cmd, str, queuemsg)
    int			cmd;		/* Command character */
    reg2 char *		str;		/* Argument string */
    int			queuemsg;	/* NZ to put out msg if no cron */
    {
    reg1 char *		buf;		/* Malloc'ed buffer for command */
    reg3 int		fifofd;		/* File descriptor for FIFO */
    reg4 int		len;		/* Length to write */
    reg5 int		pid;		/* Process ID of cron daemon */

    /*
     * Find out if there is a cron daemon running, and put its process ID
     * in 'pid' if there is.  If there is no daemon, we just return (for some
     * communication schemes), since opening the FIFO would cause us to hang.
     */
    pid = cronpid ();
#ifdef USGFIFO
    if (pid < 0)
	{
	if (queuemsg)
	    {
	    (void) fprintf (stderr, "%s: no cron daemon running, queued\n",
	      progname);
	    }
	return SEND_QUEUED;
	}
#endif /* USGFIFO */
    /*
     * We now know cron is running.  Open the FIFO and write a command to it.
     */
    fifofd = openfifo (FIFONAME, 1);	/* Open the fifo */
    if (fifofd < 0)			/* Still not open? */
	{
	(void) fprintf (stderr, "%s:  couldn't open %s: %s\n",
	  progname, FIFONAME, errmsg (0));
	return SEND_ERROR;
	}
    /*
     * Because of the way cron works, we must write the command in a single
     * call to "writefifo".  So we malloc a buffer to build stuff in.
     */
    len = strlen (str) + 3;		/* 3 is for cmd, space, \n */
    buf = malloc ((unsigned) len);
    if (buf == NULL)
	{
	(void) fprintf (stderr, "%s:  unexpectedly out of memory\n",
	  progname);
	return SEND_ERROR;
	}
    buf[0] = cmd;
    buf[1] = ' ';
    (void) strcpy (buf + 2, str);
    buf[len - 1] = '\n';
    if (writefifo (fifofd, buf, len) != len)
	{
	(void) fprintf (stderr, "%s:  couldn't write %s: %s\n",
	  progname, FIFONAME, errmsg (0));
	(void) closefifo (fifofd);
	return SEND_ERROR;
	}
    (void) closefifo (fifofd);
#ifndef BSDSOCKET
    /*
     * Inform the cron process that we have written the FIFO
     */
    if (pid < 0  ||  kill (pid, READFIFO) < 0)
	{
	(void) fprintf (stderr,
	   "%s:  trouble signaling cron (%s), queued\n", progname, errmsg (0));
	return SEND_QUEUED;
	}
#endif /* BSDSOCKET */
    return SEND_OK;
    }

static int cronpid ()		/* Return cron's process ID, or -1 if none */
    {
    reg1 FILE *		lock;	/* Lock file access */
    time_t		now;	/* The current system time */
    int			pid;	/* Pid to be read from the lock file */
    struct stat		statbuf; /* Stat buffer for learning about time log */

    lock = fopen (LOCKFILE, "r"); /* Open lock file to read PID */
    if (lock == NULL)		/* Failures are same as if cron didn't exist */
	return -1;
    /* Read PID from lock file */
    if (fscanf (lock, "%d", &pid) != 1)
	{
	(void) fclose (lock);
	return -1;
	}
    (void) fclose (lock);
    /*
     * Make sure the process really exists.
     */
    if (kill (pid, 0) < 0  &&  errno == ESRCH)
	return -1;
    /*
     * Make sure that the time-log file has been updated recently.  If
     * it hasn't, the PID probably belongs to some non-cron process.
     */
    (void) time (&now);
    if (stat (TIMELOG, &statbuf) < 0
      ||  statbuf.st_mtime < now - 2 * MINUPDLAST)
	return -1;
    return pid;
    }

void setlibdir ()		/* Change directories to LIBDIR */
    {
    /*
     * Change directories to LIBDIR.  We do this by changing to
     * CRONTABDIR and then to LIBDIR, in case LIBDIR is relative to
     * CRONTABDIR.
     */
    if (chdir (CRONTABDIR) < 0)
	{
	(void) fprintf (stderr, "%s:  trouble cd'ing to %s: %s\n",
	  progname, CRONTABDIR, errmsg (0));
	exit (1);
	}
    if (chdir (LIBDIR) < 0)
	{
	(void) fprintf (stderr, "%s:  trouble cd'ing to %s: %s\n",
	  progname, LIBDIR, errmsg (0));
	exit (1);
	}
    }

char * getname (user)			/* Get user's name */
    reg1 char *		user;		/* Original user name */
    {
    reg2 struct passwd * pwent;		/* Password entry for UID */
    reg3 int		uid;		/* User's numerical ID */

    if (user == NULL)			/* If no user name, use getuid */
	{
	uid = getuid ();
	pwent = getpwuid (uid);
	}
    else if (isdigit (*user))		/* If starts with digit, assume UID */
	{
	uid = atoi (user);
	user = NULL;			/* Force err msg below to be numeric */
	pwent = getpwuid (uid);
	}
    else
	pwent = getpwnam (user);

    endpwent ();
    userpwent = pwent;

    if (pwent == NULL)
	{
	if (user == NULL)
	    (void) fprintf (stderr,
	      "%s:  couldn't find /etc/passwd entry for UID %d\n",
	      progname, uid);
	else
	    (void) fprintf (stderr,
	      "%s:  couldn't find /etc/passwd entry for user %s\n",
	      progname, user);
	return NULL;
	}
    else
	return pwent->pw_name;		/* Return user name */
    }
